# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from sahara import context
import sahara.exceptions as ex
from sahara.i18n import _
from sahara.service.api import v10 as api
from sahara.service.health import verification_base
from sahara.service.validations import acl
import sahara.service.validations.base as b

CONF = cfg.CONF


def check_cluster_create(data, **kwargs):
    b.check_cluster_unique_name(data['name'])
    _check_cluster_create(data)


def check_multiple_clusters_create(data, **kwargs):
    _check_cluster_create(data)
    for counter in range(data['count']):
        cluster_name = api.get_multiple_cluster_name(data['count'],
                                                     data['name'],
                                                     counter + 1)
        b.check_cluster_unique_name(cluster_name)


def check_one_or_multiple_clusters_create(data, **kwargs):
    if data.get('count', None) is not None:
        check_multiple_clusters_create(data, **kwargs)
    else:
        check_cluster_create(data, **kwargs)


def _check_cluster_create(data):
    plugin_version = 'hadoop_version'
    if data.get('plugin_version'):
        plugin_version = 'plugin_version'
    b.check_plugin_name_exists(data['plugin_name'])
    b.check_plugin_supports_version(data['plugin_name'],
                                    data[plugin_version])
    b.check_plugin_labels(
        data['plugin_name'], data[plugin_version])

    if data.get('cluster_template_id'):
        ct_id = data['cluster_template_id']
        b.check_cluster_template_exists(ct_id)
        if not data.get('node_groups'):
            b.check_node_groups_in_cluster_templates(data['name'],
                                                     data['plugin_name'],
                                                     data[plugin_version],
                                                     ct_id)

    if data.get('user_keypair_id'):
        b.check_keypair_exists(data['user_keypair_id'])

    default_image_id = _get_cluster_field(data, 'default_image_id')
    if default_image_id:
        b.check_image_registered(default_image_id)
        b.check_required_image_tags(data['plugin_name'],
                                    data[plugin_version],
                                    default_image_id)
    else:
        raise ex.NotFoundException('default_image_id',
                                   _("'%s' field is not found"))

    b.check_all_configurations(data)

    if data.get('anti_affinity'):
        b.check_node_processes(data['plugin_name'], data[plugin_version],
                               data['anti_affinity'])

    if data.get('node_groups'):
        b.check_cluster_hostnames_lengths(data['name'], data['node_groups'])

    neutron_net_id = _get_cluster_field(data, 'neutron_management_network')
    if neutron_net_id:
        b.check_network_exists(neutron_net_id)
    else:
        raise ex.NotFoundException('neutron_management_network',
                                   _("'%s' field is not found"))


def _get_cluster_field(cluster, field):
    if cluster.get(field):
        return cluster[field]

    if cluster.get('cluster_template_id'):
        cluster_template = api.get_cluster_template(
            id=cluster['cluster_template_id'])

        if cluster_template.get(field):
            return cluster_template[field]

    return None


def check_cluster_delete(cluster_id, **kwargs):
    cluster = api.get_cluster(cluster_id)

    acl.check_tenant_for_delete(context.current(), cluster)
    acl.check_protected_from_delete(cluster)


def check_cluster_update(cluster_id, data, **kwargs):
    cluster = api.get_cluster(cluster_id)

    verification = verification_base.validate_verification_ops(
        cluster, data)
    acl.check_tenant_for_update(context.current(), cluster)
    if not verification:
        acl.check_protected_from_update(cluster, data)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from castellan.common.objects import passphrase
from castellan import key_manager

from sahara import context


def delete_secret(id, ctx=None):
    """delete a secret from the external key manager

    :param id: The identifier of the secret to delete
    :param ctx: The context, and associated authentication, to use with
                this operation (defaults to the current context)
    """
    if ctx is None:
        ctx = context.current()
    key_manager.API().delete(ctx, id)


def get_secret(id, ctx=None):
    """get a secret associated with an id

    :param id: The identifier of the secret to retrieve
    :param ctx: The context, and associated authentication, to use with
                this operation (defaults to the current context)
    """
    if ctx is None:
        ctx = context.current()
    key = key_manager.API().get(ctx, id)
    return key.get_encoded()


def store_secret(secret, ctx=None):
    """store a secret and return its identifier

    :param secret: The secret to store, this should be a string
    :param ctx: The context, and associated authentication, to use with
                this operation (defaults to the current context)
    """
    if ctx is None:
        ctx = context.current()
    key = passphrase.Passphrase(secret)
    return key_manager.API().store(ctx, key)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import six
import testtools

from sahara import context
from sahara.tests.unit import base
from sahara.utils import poll_utils


class FakeCluster(object):
    def __init__(self, cluster_configs):
        self.cluster_configs = cluster_configs


class FakeOption(object):
    def __init__(self, default_value, section, name):
        self.default_value = default_value
        self.name = name
        self.applicable_target = section


class TestPollUtils(base.SaharaTestCase):
    def setUp(self):
        super(TestPollUtils, self).setUp()
        context.sleep = mock.Mock()

    @mock.patch('sahara.utils.poll_utils.LOG.debug')
    def test_poll_success(self, logger):
        poll_utils.poll(**{'get_status': lambda: True,
                           'kwargs': {}, 'timeout': 5, 'sleep': 3})
        expected_call = mock.call(
            'Operation was executed successfully in timeout 5')
        self.assertEqual(1, logger.call_count)
        self.assertEqual([expected_call], logger.call_args_list)

    @mock.patch('sahara.utils.poll_utils._get_consumed')
    def test_poll_failed_first_scenario(self, get_consumed):
        get_consumed.return_value = 0
        message = ""
        try:
            poll_utils.poll(
                **{'get_status': lambda: False, 'kwargs': {},
                   'timeout': 0, 'sleep': 3})
        except Exception as e:
            message = six.text_type(e)

        if message.find('Error ID') != -1:
            message = message.split("\n")[0]
        expected_message = "'Operation' timed out after 0 second(s)"

        self.assertEqual(expected_message, message)

    @mock.patch('sahara.utils.poll_utils._get_consumed')
    def test_poll_failed_second_scenario(self, get_consumed):
        get_consumed.return_value = 0
        message = ""
        try:
            poll_utils.poll(
                **{'get_status': lambda: False, 'kwargs': {},
                   'timeout': 0, 'sleep': 3, 'timeout_name': "some timeout"})
        except Exception as e:
            message = six.text_type(e)

        if message.find('Error ID') != -1:
            message = message.split("\n")[0]
        expected_message = ("'Operation' timed out after 0 second(s) and "
                            "following timeout was violated: some timeout")

        self.assertEqual(expected_message, message)

    @mock.patch('sahara.utils.poll_utils.LOG.debug')
    @mock.patch('sahara.utils.cluster.check_cluster_exists')
    def test_plugin_poll_first_scenario(self, cluster_exists, logger):
        cluster_exists.return_value = True
        fake_get_status = mock.Mock()
        fake_get_status.side_effect = [False, False, True]
        fake_kwargs = {'kwargs': {'cat': 'tom', 'bond': 'james bond'}}
        poll_utils.plugin_option_poll(
            FakeCluster({}), fake_get_status, FakeOption(5, 'target', 'name'),
            'fake_operation', 5, **fake_kwargs)
        expected_call = mock.call('Operation with name fake_operation was '
                                  'executed successfully in timeout 5')
        self.assertEqual([expected_call], logger.call_args_list)

    @mock.patch('sahara.utils.poll_utils.LOG.debug')
    @mock.patch('sahara.utils.cluster.check_cluster_exists')
    def test_plugin_poll_second_scenario(self, cluster_exists, logger):
        cluster_exists.return_value = False
        fake_get_status = mock.Mock()
        fake_get_status.side_effect = [False, False, True]
        fake_kwargs = {'kwargs': {'cat': 'tom', 'bond': 'james bond'}}
        poll_utils.plugin_option_poll(
            FakeCluster({'target': {'name': 7}}), fake_get_status,
            FakeOption(5, 'target', 'name'),
            'fake_operation', 5, **fake_kwargs)

        expected_call = mock.call('Operation with name fake_operation was '
                                  'executed successfully in timeout 7')
        self.assertEqual([expected_call], logger.call_args_list)

    def test_poll_exception_strategy_first_scenario(self):
        fake_get_status = mock.Mock()
        fake_get_status.side_effect = [False, ValueError()]

        with testtools.ExpectedException(ValueError):
            poll_utils.poll(fake_get_status)

    def test_poll_exception_strategy_second_scenario(self):
        fake_get_status = mock.Mock()
        fake_get_status.side_effect = [False, ValueError()]
        poll_utils.poll(fake_get_status, exception_strategy='mark_as_true')
        self.assertEqual(2, fake_get_status.call_count)

    def test_poll_exception_strategy_third_scenario(self):
        fake_get_status = mock.Mock()
        fake_get_status.side_effect = [False, ValueError(), True]
        poll_utils.poll(fake_get_status, exception_strategy='mark_as_false')
        self.assertEqual(3, fake_get_status.call_count)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from keystoneauth1 import session as keystone

from sahara import exceptions as ex
from sahara.service import sessions
from sahara.tests.unit import base


class TestSessionCache(base.SaharaTestCase):

    def test_get_session(self):
        sc = sessions.SessionCache()
        session = sc.get_session()
        self.assertIsInstance(session, keystone.Session)

        self.assertRaises(ex.SaharaException,
                          sc.get_session,
                          session_type='bad service')

    @mock.patch('keystoneauth1.session.Session')
    def test_get_keystone_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='keystone')
        self.override_config('api_insecure', False, group='keystone')
        sc.get_session(sessions.SESSION_TYPE_KEYSTONE)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='keystone')
        self.override_config('api_insecure', True, group='keystone')
        sc.get_session(sessions.SESSION_TYPE_KEYSTONE)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_KEYSTONE)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_get_nova_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='nova')
        self.override_config('api_insecure', False, group='nova')
        sc.get_session(sessions.SESSION_TYPE_NOVA)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='nova')
        self.override_config('api_insecure', True, group='nova')
        sc.get_session(sessions.SESSION_TYPE_NOVA)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_NOVA)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_get_cinder_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='cinder')
        self.override_config('api_insecure', False, group='cinder')
        sc.get_session(sessions.SESSION_TYPE_CINDER)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='cinder')
        self.override_config('api_insecure', True, group='cinder')
        sc.get_session(sessions.SESSION_TYPE_CINDER)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_CINDER)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_get_neutron_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='neutron')
        self.override_config('api_insecure', False, group='neutron')
        sc.get_session(sessions.SESSION_TYPE_NEUTRON)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='neutron')
        self.override_config('api_insecure', True, group='neutron')
        sc.get_session(sessions.SESSION_TYPE_NEUTRON)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_NEUTRON)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_get_glance_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='glance')
        self.override_config('api_insecure', False, group='glance')
        sc.get_session(sessions.SESSION_TYPE_GLANCE)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='glance')
        self.override_config('api_insecure', True, group='glance')
        sc.get_session(sessions.SESSION_TYPE_GLANCE)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_GLANCE)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_get_heat_session(self, keystone_session):
        sc = sessions.SessionCache()
        self.override_config('ca_file', '/some/cacert', group='heat')
        self.override_config('api_insecure', False, group='heat')
        sc.get_session(sessions.SESSION_TYPE_HEAT)
        keystone_session.assert_called_once_with(verify='/some/cacert')

        sc = sessions.SessionCache()
        keystone_session.reset_mock()
        self.override_config('ca_file', None, group='heat')
        self.override_config('api_insecure', True, group='heat')
        sc.get_session(sessions.SESSION_TYPE_HEAT)
        keystone_session.assert_called_once_with(verify=False)

        keystone_session.reset_mock()
        sc.get_session(sessions.SESSION_TYPE_HEAT)
        self.assertFalse(keystone_session.called)

    @mock.patch('keystoneauth1.session.Session')
    def test_insecure_session(self, session):
        sc = sessions.SessionCache()
        sc.get_session(sessions.SESSION_TYPE_INSECURE)
        session.assert_called_once_with(verify=False)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara import context
from sahara.plugins import kerberos as krb
from sahara.tests.unit import base

ADD_PRINCIPAL_SCRIPT = """#!/bin/bash

mkdir -p /tmp/sahara-kerberos/

kadmin -p sahara/admin <<EOF
strong
addprinc -randkey bond/sahara-in.loc@SK
xst -k /tmp/sahara-kerberos/bond-sahara-in.loc.keytab bond/sahara-in.loc@SK
exit
EOF

sudo chown bond:bond /tmp/sahara-kerberos/bond-sahara-in.loc.keytab
"""
CRON_FILE = """# Once per hour refreshes tickets for user bond
0 * * * * root /tmp/sahara-kerberos/e1.sh
"""
CRON_SCRIPT = ('#!/bin/bash\n\n'
               'sudo -u bond kinit -p bond/sahara-in.loc@SK -kt '
               '/tmp/sahara-kerberos/bond-sahara-in.loc.keytab\n')


class FakeObject(dict):
    def __init__(self, dict):
        self._dict = dict
        super(FakeObject, self).__init__(**dict)

    def to_dict(self):
        return self._dict


class TestKerberosBase(base.SaharaTestCase):
    def test_list_configs(self):
        self.assertEqual(7, len(krb.get_config_list()))

    @mock.patch('sahara.conductor.API.cluster_update')
    @mock.patch('sahara.conductor.API.cluster_get')
    @mock.patch('sahara.service.castellan.utils.store_secret')
    @mock.patch('sahara.service.castellan.utils.get_secret')
    def test_get_server_password(
            self, get_secret, store_secret, cluster_get_mock,
            cluster_update_mock):
        cl = mock.Mock(
            node_groups=[], cluster_configs={}, extra={})
        ctx = context.ctx()
        cluster_get_mock.return_value = cl
        store_secret.return_value = 'secret-id'
        krb.get_server_password(cl)

        self.assertEqual(1, cluster_get_mock.call_count)
        self.assertEqual(1, cluster_update_mock.call_count)
        self.assertEqual([
            mock.call(ctx, cl, {'extra': {'admin-passwd-kdc': 'secret-id'}})],
            cluster_update_mock.call_args_list)

        self.assertEqual(1, get_secret.call_count)
        self.assertEqual(1, store_secret.call_count)

        cl = mock.Mock(
            node_groups=[], cluster_configs={},
            extra=FakeObject({'admin-passwd-kdc': 'secret-id'}))
        cluster_get_mock.return_value = cl
        krb.get_server_password(cl)

        self.assertEqual(2, get_secret.call_count)
        self.assertEqual(1, store_secret.call_count)
        self.assertEqual(1, cluster_update_mock.call_count)

        cl = mock.Mock(
            node_groups=[], cluster_configs=FakeObject({
                'Existing KDC': True, 'Admin password': 'THE BEST EVER'}),
            extra=FakeObject({'admin-passwd-kdc': 'secret-id'}))
        cluster_get_mock.return_value = cl
        get_secret.return_value = 'THE BEST EVER'
        self.assertEqual('THE BEST EVER', krb.get_server_password(cl))

    def test_base_configs(self):
        cluster = mock.Mock(
            cluster_configs={'Kerberos': {'Enable Kerberos Security': True}},
            node_groups=[],
        )
        self.assertTrue(krb.is_kerberos_security_enabled(cluster))

        cluster = mock.Mock(
            cluster_configs={'Kerberos': {'Enable Kerberos Security': False}},
            node_groups=[],
        )
        self.assertFalse(krb.is_kerberos_security_enabled(cluster))

    @mock.patch('sahara.plugins.kerberos.get_server_password')
    def test_get_server_installation_script(self, get):
        cluster = mock.Mock(node_groups=[], cluster_configs={})
        get.return_value = 'password'
        krb._get_server_installation_script(
            cluster, 'server.novalocal', 'centos', '6.7')

    @mock.patch('sahara.plugins.kerberos.get_realm_name')
    @mock.patch('sahara.plugins.kerberos._get_short_uuid')
    @mock.patch('sahara.plugins.kerberos.get_server_password')
    def test_create_keytabs_for_user(self, get_password, get_uuid, realm):
        get_uuid.return_value = 'e1'
        realm.return_value = 'SK'
        get_password.return_value = "strong"
        cluster = mock.Mock(node_groups=[], cluster_configs={})
        instance = mock.Mock()
        instance.fqdn = mock.Mock()
        instance.fqdn.return_value = "in.loc"
        data = krb._get_script_for_user_creation(cluster, instance, 'bond')
        self.assertEqual(ADD_PRINCIPAL_SCRIPT, data[0])
        self.assertEqual(CRON_FILE, data[1])
        self.assertEqual(CRON_SCRIPT, data[2])
        self.assertEqual('/tmp/sahara-kerberos/e1.sh', data[3])



# Copyright (c) 2013 Mirantis Inc.
# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from oslo_config import cfg
import six

from sahara import exceptions as ex
from sahara.i18n import _

# These options are for SSH remote only
ssh_opts = [
    cfg.IntOpt('global_remote_threshold', default=100,
               help='Maximum number of remote operations that will '
                    'be running at the same time. Note that each '
                    'remote operation requires its own process to '
                    'run.'),
    cfg.IntOpt('cluster_remote_threshold', default=70,
               help='The same as global_remote_threshold, but for '
                    'a single cluster.'),
    cfg.StrOpt('proxy_command', default='',
               help='Proxy command used to connect to instances. If set, this '
               'command should open a netcat socket, that Sahara will use for '
               'SSH and HTTP connections. Use {host} and {port} to describe '
               'the destination. Other available keywords: {tenant_id}, '
               '{network_id}, {router_id}.'),
    cfg.BoolOpt('proxy_command_use_internal_ip', default=False,
                help='Force proxy_command usage to be consuming internal IP '
                'always, instead of management IP. Ignored if proxy_command '
                'is not set.')
]


CONF = cfg.CONF
CONF.register_opts(ssh_opts)


DRIVER = None


@six.add_metaclass(abc.ABCMeta)
class RemoteDriver(object):
    @abc.abstractmethod
    def setup_remote(self, engine):
        """Performs driver initialization."""

    @abc.abstractmethod
    def get_remote(self, instance):
        """Returns driver specific Remote."""

    @abc.abstractmethod
    def get_userdata_template(self):
        """Returns userdata template preparing instance to work with driver."""

    @abc.abstractmethod
    def get_type_and_version(self):
        """Returns engine type and version

         Result should be in the form 'type.major.minor'.
         """


@six.add_metaclass(abc.ABCMeta)
class TerminalOnlyRemote(object):

    @abc.abstractmethod
    def execute_command(self, cmd, run_as_root=False, get_stderr=False,
                        raise_when_error=True, timeout=300):
        """Execute specified command remotely using existing ssh connection.

        Return exit code, stdout data and stderr data of the executed command.
        """

    @abc.abstractmethod
    def get_os_distrib(self):
        """Returns the OS distribution running on the target machine."""


@six.add_metaclass(abc.ABCMeta)
class Remote(TerminalOnlyRemote):

    @abc.abstractmethod
    def get_neutron_info(self):
        """Returns dict which later could be passed to get_http_client."""

    @abc.abstractmethod
    def get_http_client(self, port, info=None):
        """Returns HTTP client for a given instance's port."""

    @abc.abstractmethod
    def close_http_session(self, port):
        """Closes cached HTTP session for a given instance's port."""

    @abc.abstractmethod
    def write_file_to(self, remote_file, data, run_as_root=False, timeout=120):
        """Create remote file and write the given data to it.

        Uses existing ssh connection.
        """

    @abc.abstractmethod
    def append_to_file(self, r_file, data, run_as_root=False, timeout=120):
        """Append the given data to remote file.

        Uses existing ssh connection.
        """

    @abc.abstractmethod
    def write_files_to(self, files, run_as_root=False, timeout=120):
        """Copy file->data dictionary in a single ssh connection."""

    @abc.abstractmethod
    def append_to_files(self, files, run_as_root=False, timeout=120):
        """Copy file->data dictionary in a single ssh connection."""

    @abc.abstractmethod
    def read_file_from(self, remote_file, run_as_root=False, timeout=120):
        """Read remote file from the specified host and return given data."""

    @abc.abstractmethod
    def replace_remote_string(self, remote_file, old_str, new_str,
                              timeout=120):
        """Replaces strings in remote file using sed command."""


def setup_remote(driver, engine):
    global DRIVER

    DRIVER = driver
    DRIVER.setup_remote(engine)


def get_remote_type_and_version():
    return DRIVER.get_type_and_version()


def _check_driver_is_loaded():
    if not DRIVER:
        raise ex.SystemError(_('Remote driver is not loaded. Most probably '
                               'you see this error because you are running '
                               'Sahara in distributed mode and it is broken.'
                               'Try running sahara-all instead.'))


def get_remote(instance):
    """Returns Remote for a given instance."""
    _check_driver_is_loaded()
    return DRIVER.get_remote(instance)


def get_userdata_template():
    """Returns userdata template as a string."""
    _check_driver_is_loaded()
    return DRIVER.get_userdata_template()



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_utils import uuidutils

from sahara.service.edp.data_sources.base import DataSourceType

import testtools


class DataSourceBaseTestCase(testtools.TestCase):
    def setUp(self):
        super(DataSourceBaseTestCase, self).setUp()
        self.ds_base = DataSourceType()

    def test_construct_url_no_placeholders(self):
        base_url = "swift://container/input"
        job_exec_id = uuidutils.generate_uuid()

        url = self.ds_base.construct_url(base_url, job_exec_id)

        self.assertEqual(base_url, url)

    def test_construct_url_job_exec_id_placeholder(self):
        base_url = "swift://container/input.%JOB_EXEC_ID%.out"
        job_exec_id = uuidutils.generate_uuid()

        url = self.ds_base.construct_url(base_url, job_exec_id)

        self.assertEqual(
            "swift://container/input." + job_exec_id + ".out", url)

    def test_construct_url_randstr_placeholder(self):
        base_url = "swift://container/input.%RANDSTR(4)%.%RANDSTR(7)%.out"
        job_exec_id = uuidutils.generate_uuid()

        url = self.ds_base.construct_url(base_url, job_exec_id)

        self.assertRegex(
            url, "swift://container/input\.[a-z]{4}\.[a-z]{7}\.out")

    def test_construct_url_randstr_and_job_exec_id_placeholder(self):
        base_url = "swift://container/input.%JOB_EXEC_ID%.%RANDSTR(7)%.out"
        job_exec_id = uuidutils.generate_uuid()

        url = self.ds_base.construct_url(base_url, job_exec_id)

        self.assertRegex(
            url, "swift://container/input." + job_exec_id + "\.[a-z]{7}\.out")

    def test_get_urls(self):
        url = 'test://url'
        cluster = mock.Mock()
        job_exec_id = 'test_id'

        self.assertEqual((url, url), self.ds_base.get_urls(url,
                         cluster, job_exec_id))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sahara import exceptions as ex
from sahara.i18n import _
from sahara.service.api import v10 as api
import sahara.service.validations.base as b
from sahara.service.validations import shares


def check_cluster_template_create(data, **kwargs):
    plugin_version = 'hadoop_version'
    if data.get('plugin_version'):
        plugin_version = 'plugin_version'

    b.check_cluster_template_unique_name(data['name'])
    b.check_plugin_name_exists(data['plugin_name'])
    b.check_plugin_supports_version(data['plugin_name'],
                                    data[plugin_version])

    if data.get('default_image_id'):
        b.check_image_registered(data['default_image_id'])
        b.check_required_image_tags(data['plugin_name'],
                                    data[plugin_version],
                                    data['default_image_id'])

    b.check_all_configurations(data)

    if data.get('anti_affinity'):
        b.check_node_processes(data['plugin_name'], data[plugin_version],
                               data['anti_affinity'])

    if data.get('neutron_management_network'):
        b.check_network_exists(data['neutron_management_network'])

    if data.get('shares'):
        shares.check_shares(data['shares'])


def check_cluster_template_usage(cluster_template_id, **kwargs):
    users = []

    for cluster in api.get_clusters():
        if cluster_template_id == cluster.cluster_template_id:
            users.append(cluster.name)

    if users:
        raise ex.InvalidReferenceException(
            _("Cluster template %(id)s in use by %(clusters)s") %
            {'id': cluster_template_id,
             'clusters': ', '.join(users)})


def check_cluster_template_update(cluster_template_id, data, **kwargs):
    plugin_version = 'hadoop_version'
    if data.get('plugin_version'):
        plugin_version = 'plugin_version'

    if data.get('plugin_name') and not data.get(plugin_version):
        raise ex.InvalidReferenceException(
            _("You must specify a %s value "
              "for your plugin_name") % plugin_version)

    if data.get('plugin_name'):
        plugin = data['plugin_name']
        version = data[plugin_version]
        b.check_plugin_name_exists(plugin)
        b.check_plugin_supports_version(plugin, version)
        b.check_all_configurations(data)
    else:
        cluster_template = api.get_cluster_template(cluster_template_id)
        plugin = cluster_template.plugin_name
        if data.get(plugin_version):
            version = data.get(plugin_version)
            b.check_plugin_supports_version(plugin, version)
        else:
            version = cluster_template.hadoop_version

    if data.get('default_image_id'):
        b.check_image_registered(data['default_image_id'])
        b.check_required_image_tags(plugin, version, data['default_image_id'])

    if data.get('anti_affinity'):
        b.check_node_processes(plugin, version, data['anti_affinity'])

    if data.get('neutron_management_network'):
        b.check_network_exists(data['neutron_management_network'])

    if data.get('shares'):
        shares.check_shares(data['shares'])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
import six
import yaml

from sahara.plugins import provisioning as plugin_provisioning
from sahara.service.heat import commons as heat_common
from sahara.utils import cluster as cl
from sahara.utils import general as g
from sahara.utils.openstack import base as b
from sahara.utils.openstack import heat as h
from sahara.utils.openstack import neutron
from sahara.utils.openstack import nova

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

SSH_PORT = 22
INSTANCE_RESOURCE_NAME = "inst"
SERVER_GROUP_NAMES = "servgroups"
AUTO_SECURITY_GROUP_PARAM_NAME = "autosecgroup"
INTERNAL_DESIGNATE_REC = "internal_designate_record"
INTERNAL_DESIGNATE_REV_REC = "internal_designate_reverse_record"
EXTERNAL_DESIGNATE_REC = "external_designate_record"
EXTERNAL_DESIGNATE_REV_REC = "external_designate_reverse_record"

# TODO(vgridnev): Using insecure flag until correct way to pass certificate
# will be invented
WAIT_CONDITION_SCRIPT_TEMPLATE = '''
while true; do
    wc_notify --insecure --data-binary '{"status": "SUCCESS"}'
    if [ $? -eq 0 ]; then
        break
    fi
    sleep 10
done
'''

heat_engine_opts = [
    cfg.BoolOpt(
        'heat_enable_wait_condition', default=True,
        help="Enable wait condition feature to reduce polling during cluster "
             "creation")
]
CONF.register_opts(heat_engine_opts)


def _get_inst_name(ng):
    return {
        "list_join": [
            '-',
            [ng.cluster.name.lower(), ng.name.lower(),
             {"get_param": "instance_index"}]
        ]
    }


def _get_inst_domain_name(domain):
    return {
        "list_join": [
            '.',
            [{"get_attr": [INSTANCE_RESOURCE_NAME, "name"]}, domain]]
    }


def _get_aa_group_name(cluster, server_group_index):
    return g.generate_aa_group_name(cluster.name, server_group_index)


def _get_port_name(ng):
    return {
        "list_join": [
            '-',
            [ng.cluster.name.lower(), ng.name.lower(),
             {"get_param": "instance_index"},
             "port"]
        ]
    }


def _get_floating_name(ng):
    return {
        "list_join": [
            '-',
            [ng.cluster.name.lower(), ng.name.lower(),
             {"get_param": "instance_index"},
             "floating"]
        ]
    }


def _get_floating_assoc_name(ng):
    return {
        "list_join": [
            '-',
            [ng.cluster.name.lower(), ng.name.lower(),
             {"get_param": "instance_index"},
             "floating", "assoc"]
        ]
    }


def _get_volume_name(ng):
    return {
        "list_join": [
            '-',
            [ng.cluster.name.lower(), ng.name.lower(),
             {"get_param": "instance_index"},
             "volume", {"get_param": "volume_index"}]
        ]
    }


def _get_wc_handle_name(inst_name):
    return '%s-wc-handle' % inst_name


def _get_wc_waiter_name(inst_name):
    return '%s-wc-waiter' % inst_name


def _get_index_from_inst_name(inst_name):
    return inst_name.split('-')[-1]


class ClusterStack(object):
    def __init__(self, cluster):
        self.cluster = cluster
        self.node_groups_extra = {}
        self.files = {}
        self.last_updated_time = None
        self.base_info = (
            "Data Processing Cluster by Sahara\n"
            "Sahara cluster name: {cluster}\n"
            "Sahara engine: {version}".format(
                cluster=cluster.name, version=heat_common.HEAT_ENGINE_VERSION)
        )
        self._current_sg_index = 1

    def _node_group_description(self, ng):
        return "{info}\nNode group {node_group}".format(
            info=self.base_info, node_group=ng.name)

    def _asg_for_node_group_description(self, ng):
        return ("{info}\n"
                "Auto security group for Sahara Node Group: "
                "{node_group}".format(info=self.base_info, node_group=ng.name))

    def _volume_for_node_group_description(self, ng):
        return ("{info}\n"
                "Volume for Sahara Node Group {node_group}".format(
                    node_group=ng.name, info=self.base_info))

    def add_node_group_extra(self, node_group_id, node_count,
                             gen_userdata_func, instances_to_delete=None):
        self.node_groups_extra[node_group_id] = {
            'node_count': node_count,
            'gen_userdata_func': gen_userdata_func,
            'instances_to_delete': instances_to_delete
        }

    def _get_main_template(self, instances_to_delete=None):
        outputs = {}
        resources = self._serialize_resources(outputs, instances_to_delete)
        return yaml.safe_dump({
            "heat_template_version": heat_common.HEAT_TEMPLATE_VERSION,
            "description": self.base_info,
            "resources": resources,
            "outputs": outputs
        })

    def instantiate(self, update_existing, disable_rollback=True,
                    instances_to_delete=None):
        main_tmpl = self._get_main_template(instances_to_delete)
        kwargs = {
            'stack_name': self.cluster.stack_name,
            'timeout_mins': 180,
            'disable_rollback': disable_rollback,
            'parameters': {},
            'template': main_tmpl,
            'files': self.files
        }

        if CONF.heat_stack_tags:
            kwargs['tags'] = ",".join(CONF.heat_stack_tags)

        log_kwargs = copy.deepcopy(kwargs)
        log_kwargs['template'] = yaml.safe_load(log_kwargs['template'])
        for filename in log_kwargs['files'].keys():
            log_kwargs['files'][filename] = yaml.safe_load(
                log_kwargs['files'][filename])
        log_kwargs = json.dumps(log_kwargs)

        if not update_existing:
            LOG.debug("Creating Heat stack with args: \n{args}"
                      .format(args=log_kwargs))
            b.execute_with_retries(h.client().stacks.create, **kwargs)
        else:
            stack = h.get_stack(self.cluster.stack_name)
            self.last_updated_time = stack.updated_time
            LOG.debug("Updating Heat stack {stack} with args: \n"
                      "{args}".format(stack=stack, args=log_kwargs))
            b.execute_with_retries(stack.update, **kwargs)

    def _get_server_group_name(self):
        index = self._current_sg_index
        # computing server group index in round robin fashion
        if index < self.cluster.anti_affinity_ratio:
            self._current_sg_index = (index + 1)
        else:
            self._current_sg_index = 1
        return _get_aa_group_name(self.cluster, self._current_sg_index)

    def _need_aa_server_group(self, node_group):
        for node_process in node_group.node_processes:
            if node_process in self.cluster.anti_affinity:
                return True
        return False

    def _get_anti_affinity_scheduler_hints(self, node_group):
        if not self._need_aa_server_group(node_group):
            return {}

        return {
            "scheduler_hints": {
                "group": {
                    "get_param": [SERVER_GROUP_NAMES,
                                  {"get_param": "instance_index"}]
                }
            }
        }

    def _serialize_resources(self, outputs, instances_to_delete=None):
        resources = {}

        if self.cluster.anti_affinity:
            # Creating server groups equal to the anti_affinity_ratio
            for i in range(0, self.cluster.anti_affinity_ratio):
                resources.update(self._serialize_aa_server_group(i + 1))

        for ng in self.cluster.node_groups:
            resources.update(self._serialize_ng_group(ng, outputs,
                                                      instances_to_delete))

        for ng in self.cluster.node_groups:
            resources.update(self._serialize_auto_security_group(ng))

        return resources

    def _serialize_ng_group(self, ng, outputs, instances_to_delete=None):
        ng_file_name = "file://" + ng.name + ".yaml"
        self.files[ng_file_name] = self._serialize_ng_file(ng)

        outputs[ng.name + "-instances"] = {
            "value": {"get_attr": [ng.name, "instance"]}}
        properties = {"instance_index": "%index%"}

        if ng.cluster.anti_affinity:
            ng_count = self.node_groups_extra[ng.id]['node_count']
            # assuming instance_index also start from index 0
            for i in range(0, ng_count):
                server_group_name = self._get_server_group_name()
                server_group_resource = {
                    "get_resource": server_group_name
                }
                if SERVER_GROUP_NAMES not in properties:
                    properties[SERVER_GROUP_NAMES] = []

                properties[SERVER_GROUP_NAMES].insert(i, server_group_resource)

        if ng.auto_security_group:
            properties[AUTO_SECURITY_GROUP_PARAM_NAME] = {
                'get_resource': g.generate_auto_security_group_name(ng)}

        removal_policies = []
        if self.node_groups_extra[ng.id]['instances_to_delete']:
            resource_list = []
            for name in self.node_groups_extra[ng.id]['instances_to_delete']:
                resource_list.append(_get_index_from_inst_name(name))
            removal_policies.append({'resource_list': resource_list})

        return {
            ng.name: {
                "type": "OS::Heat::ResourceGroup",
                "properties": {
                    "count": self.node_groups_extra[ng.id]['node_count'],
                    "removal_policies": removal_policies,
                    "resource_def": {
                        "type": ng_file_name,
                        "properties": properties
                    }
                }
            }
        }

    def _serialize_ng_file(self, ng):
        parameters = {"instance_index": {"type": "string"}}

        if ng.cluster.anti_affinity:
            parameters[SERVER_GROUP_NAMES] = {"type": "comma_delimited_list",
                                              "default": []}

        if ng.auto_security_group:
            parameters[AUTO_SECURITY_GROUP_PARAM_NAME] = {'type': "string"}

        return yaml.safe_dump({
            "heat_template_version": heat_common.HEAT_TEMPLATE_VERSION,
            "description": self._node_group_description(ng),
            "parameters": parameters,
            "resources": self._serialize_instance(ng),
            "outputs": {
                "instance": {"value": {
                    "physical_id": {"get_resource": INSTANCE_RESOURCE_NAME},
                    "name": {"get_attr": [INSTANCE_RESOURCE_NAME, "name"]}
                }}}
        })

    def _serialize_auto_security_group(self, ng):
        if not ng.auto_security_group:
            return {}
        security_group_name = g.generate_auto_security_group_name(ng)
        security_group_description = self._asg_for_node_group_description(ng)

        res_type = "OS::Neutron::SecurityGroup"
        desc_key = "description"
        rules_key = "rules"
        create_rule = lambda ip_version, cidr, proto, from_port, to_port: {
            "ethertype": "IPv{}".format(ip_version),
            "remote_ip_prefix": cidr,
            "protocol": proto,
            "port_range_min": six.text_type(from_port),
            "port_range_max": six.text_type(to_port)}

        rules = self._serialize_auto_security_group_rules(ng, create_rule)

        return {
            security_group_name: {
                "type": res_type,
                "properties": {
                    desc_key: security_group_description,
                    rules_key: rules
                }
            }
        }

    def _serialize_auto_security_group_rules(self, ng, create_rule):
        rules = []
        for port in ng.open_ports:
            rules.append(create_rule(4, '0.0.0.0/0', 'tcp', port, port))
            rules.append(create_rule(6, '::/0', 'tcp', port, port))

        rules.append(create_rule(4, '0.0.0.0/0', 'tcp', SSH_PORT, SSH_PORT))
        rules.append(create_rule(6, '::/0', 'tcp', SSH_PORT, SSH_PORT))

        # open all traffic for private networks
        for cidr in neutron.get_private_network_cidrs(ng.cluster):
            ip_ver = 6 if ':' in cidr else 4
            for protocol in ['tcp', 'udp']:
                rules.append(create_rule(ip_ver, cidr, protocol, 1, 65535))
            rules.append(create_rule(ip_ver, cidr, 'icmp', 0, 255))

        return rules

    @staticmethod
    def _get_wait_condition_timeout(ng):
        configs = ng.cluster.cluster_configs
        timeout_cfg = plugin_provisioning.HEAT_WAIT_CONDITION_TIMEOUT
        cfg_target = timeout_cfg.applicable_target
        cfg_name = timeout_cfg.name
        return int(configs.get(cfg_target,
                               {}).get(cfg_name, timeout_cfg.default_value))

    def _serialize_designate_records(self):
        if not self.cluster.use_designate_feature():
            return {}
        hostname = _get_inst_domain_name(self.cluster.domain_name)
        return {
            INTERNAL_DESIGNATE_REC: {
                'type': 'OS::Designate::Record',
                'properties': {
                    'name': hostname,
                    'type': 'A',
                    'data': {'get_attr': [
                        INSTANCE_RESOURCE_NAME, 'networks', 'private', 0]},
                    'domain': self.cluster.domain_name
                }
            },
            EXTERNAL_DESIGNATE_REC: {
                'type': 'OS::Designate::Record',
                'properties': {
                    'name': hostname,
                    'type': 'A',
                    'data': {'get_attr': ['floating_ip', 'ip']},
                    'domain': self.cluster.domain_name
                }
            }
        }

    def _serialize_designate_reverse_records(self):

        if not self.cluster.use_designate_feature():
            return {}

        def _generate_reversed_ip(ip):
            return {
                'list_join': [
                    '.',
                    [
                        {'str_split': ['.', ip, 3]},
                        {'str_split': ['.', ip, 2]},
                        {'str_split': ['.', ip, 1]},
                        {'str_split': ['.', ip, 0]},
                        'in-addr.arpa.'
                    ]
                ]
            }

        hostname = _get_inst_domain_name(self.cluster.domain_name)
        return {
            INTERNAL_DESIGNATE_REV_REC: {
                'type': 'OS::Designate::Record',
                'properties': {
                    'name': _generate_reversed_ip({'get_attr': [
                        INSTANCE_RESOURCE_NAME, 'networks', 'private', 0]}),
                    'type': 'PTR',
                    'data': hostname,
                    'domain': 'in-addr.arpa.'
                }
            },
            EXTERNAL_DESIGNATE_REV_REC: {
                'type': 'OS::Designate::Record',
                'properties': {
                    'name': _generate_reversed_ip(
                        {'get_attr': ['floating_ip', 'ip']}),
                    'type': 'PTR',
                    'data': hostname,
                    'domain': 'in-addr.arpa.'
                }
            }
        }

    def _serialize_instance(self, ng):
        resources = {}
        properties = {}

        inst_name = _get_inst_name(ng)
        private_net = self.cluster.neutron_management_network

        sec_groups = self._get_security_groups(ng)

        # Check if cluster contains user key-pair and include it to template.
        if self.cluster.user_keypair_id:
            properties["key_name"] = self.cluster.user_keypair_id

        port_name = _get_port_name(ng)

        resources.update(self._serialize_port(
            port_name, private_net, sec_groups))

        properties["networks"] = [{"port": {"get_resource": "port"}}]

        if ng.floating_ip_pool:
            resources.update(self._serialize_neutron_floating(ng))

        gen_userdata_func = self.node_groups_extra[ng.id]['gen_userdata_func']
        key_script = gen_userdata_func(ng, inst_name)
        if CONF.heat_enable_wait_condition:
            etc_hosts = cl.etc_hosts_entry_for_service('orchestration')
            if etc_hosts:
                etc_hosts = "echo '%s' | sudo tee -a /etc/hosts" % etc_hosts
            tml = [key_script, WAIT_CONDITION_SCRIPT_TEMPLATE]
            if etc_hosts:
                tml = [key_script, etc_hosts, WAIT_CONDITION_SCRIPT_TEMPLATE]
            userdata = {
                "str_replace": {
                    "template": "\n".join(tml),
                    "params": {
                        "wc_notify": {
                            "get_attr": [
                                _get_wc_handle_name(ng.name),
                                "curl_cli"
                            ]
                        }
                    }
                }
            }
        else:
            userdata = key_script

        if ng.availability_zone:
            properties["availability_zone"] = ng.availability_zone

        properties.update(self._get_anti_affinity_scheduler_hints(ng))

        properties.update({
            "name": inst_name,
            "flavor": six.text_type(ng.flavor_id),
            "admin_user": ng.image_username,
            "user_data": userdata
        })

        if ng.boot_from_volume:
            resources.update(self._get_bootable_volume(ng))
            properties["block_device_mapping"] = [
                {"device_name": "vda",
                 "volume_id": {"get_resource": "bootable_volume"},
                 "delete_on_termination": "true"}]

        else:
            properties.update({"image": ng.get_image_id()})

        resources.update({
            INSTANCE_RESOURCE_NAME: {
                "type": "OS::Nova::Server",
                "properties": properties
            }
        })

        resources.update(self._serialize_designate_records())
        resources.update(self._serialize_designate_reverse_records())
        resources.update(self._serialize_volume(ng))
        resources.update(self._serialize_wait_condition(ng))
        return resources

    def _get_bootable_volume(self, node_group):
        node_group_flavor = nova.get_flavor(id=node_group.flavor_id)
        image_size = node_group_flavor.disk

        properties = {}
        properties["size"] = image_size
        properties["image"] = node_group.get_image_id()

        if node_group.boot_volume_type:
            properties["volume_type"] = node_group.boot_volume_type

        if node_group.boot_volume_availability_zone:
            properties["availability_zone"] = (
                node_group.boot_volume_availability_zone
            )

        if node_group.boot_volume_local_to_instance:
            properties["scheduler_hints"] = {
                "local_to_instance": {"get_param": "instance"}}

        return {
            "bootable_volume": {
                "type": "OS::Cinder::Volume",
                "properties": properties
            }
        }

    def _serialize_wait_condition(self, ng):
        if not CONF.heat_enable_wait_condition:
            return {}
        return {
            _get_wc_handle_name(ng.name): {
                "type": "OS::Heat::WaitConditionHandle"
            },
            _get_wc_waiter_name(ng.name): {
                "type": "OS::Heat::WaitCondition",
                "depends_on": INSTANCE_RESOURCE_NAME,
                "properties": {
                    "timeout": self._get_wait_condition_timeout(ng),
                    "handle": {"get_resource": _get_wc_handle_name(ng.name)}
                }
            }
        }

    def _serialize_neutron_floating(self, ng):
        return {
            "floating_ip": {
                "type": "OS::Neutron::FloatingIP",
                "properties": {
                    "floating_network_id": ng.floating_ip_pool,
                    "port_id": {"get_resource": "port"}
                }
            }
        }

    def _serialize_port(self, port_name, fixed_net_id, security_groups):
        properties = {
            "network_id": fixed_net_id,
            "replacement_policy": "AUTO",
            "name": port_name
        }
        if security_groups:
            properties["security_groups"] = security_groups

        return {
            "port": {
                "type": "OS::Neutron::Port",
                "properties": properties,
            }
        }

    def _serialize_volume(self, ng):
        if not ng.volumes_size or not ng.volumes_per_node:
            return {}
        volume_file_name = "file://" + ng.name + "-volume.yaml"
        self.files[volume_file_name] = self._serialize_volume_file(ng)

        return {
            ng.name: {
                "type": "OS::Heat::ResourceGroup",
                "properties": {
                    "count": ng.volumes_per_node,
                    "resource_def": {
                        "type": volume_file_name,
                        "properties": {
                            "volume_index": "%index%",
                            "instance_index": {"get_param": "instance_index"},
                            "instance": {"get_resource":
                                         INSTANCE_RESOURCE_NAME}}
                    }
                }
            }
        }

    def _serialize_volume_file(self, ng):
        volume_name = _get_volume_name(ng)
        properties = {
            "name": volume_name,
            "size": six.text_type(ng.volumes_size)
        }
        if ng.volume_type:
            properties["volume_type"] = ng.volume_type

        if ng.volumes_availability_zone:
            properties["availability_zone"] = ng.volumes_availability_zone

        if ng.volume_local_to_instance:
            properties["scheduler_hints"] = {
                "local_to_instance": {"get_param": "instance"}}

        return yaml.safe_dump({
            "heat_template_version": heat_common.HEAT_TEMPLATE_VERSION,
            "description": self._volume_for_node_group_description(ng),
            "parameters": {
                "volume_index": {
                    "type": "string"
                },
                "instance_index": {
                    "type": "string"
                },
                "instance": {
                    "type": "string"
                }},
            "resources": {
                "volume": {
                    "type": "OS::Cinder::Volume",
                    "properties": properties
                },
                "volume-attachment": {
                    "type": "OS::Cinder::VolumeAttachment",
                    "properties": {
                        "instance_uuid": {"get_param": "instance"},
                        "volume_id": {"get_resource": "volume"},
                    }
                }},
            "outputs": {}
        })

    def _get_security_groups(self, node_group):
        node_group_sg = list(node_group.security_groups or [])
        if node_group.auto_security_group:
            node_group_sg += [
                {"get_param": AUTO_SECURITY_GROUP_PARAM_NAME}
            ]
        return node_group_sg

    def _serialize_aa_server_group(self, server_group_index):
        server_group_name = _get_aa_group_name(self.cluster,
                                               server_group_index)
        return {
            server_group_name: {
                "type": "OS::Nova::ServerGroup",
                "properties": {
                    "name": server_group_name,
                    "policies": ["anti-affinity"]
                }
            }
        }

    def get_node_group_instances(self, node_group):
        cluster = node_group.cluster
        outputs = h.get_stack_outputs(cluster)
        for output in outputs:
            if output['output_key'] == node_group.name + "-instances":
                return output["output_value"]

        return []



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara.conductor import resource as r
from sahara.service.heat import templates as h
from sahara.tests.unit import base
from sahara.tests.unit import testutils as tu


class BaseTestClusterTemplate(base.SaharaWithDbTestCase):
    """Checks valid structure of Resources section in generated Heat templates.

    1. It checks templates generation with OpenStack network
    installation: Neutron.
    2. Cinder volume attachments.
    3. Basic instances creations with multi line user data provided.
    4. Anti-affinity feature with proper nova scheduler hints included
    into Heat templates.
    """
    def _make_node_groups(self, floating_ip_pool=None, volume_type=None):
        ng1 = tu.make_ng_dict('master', 42, ['namenode'], 1,
                              floating_ip_pool=floating_ip_pool, image_id=None,
                              volumes_per_node=0, volumes_size=0, id="1",
                              image_username='root', volume_type=None,
                              boot_from_volume=False, auto_security_group=True)
        ng2 = tu.make_ng_dict('worker', 42, ['datanode'], 1,
                              floating_ip_pool=floating_ip_pool, image_id=None,
                              volumes_per_node=2, volumes_size=10, id="2",
                              image_username='root', volume_type=volume_type,
                              boot_from_volume=False, auto_security_group=True)
        return ng1, ng2

    def _make_cluster(self, mng_network, ng1, ng2, anti_affinity=None,
                      domain_name=None):
        return tu.create_cluster("cluster", "tenant1", "general",
                                 "2.6.0", [ng1, ng2],
                                 user_keypair_id='user_key',
                                 neutron_management_network=mng_network,
                                 default_image_id='1', image_id=None,
                                 anti_affinity=anti_affinity or [],
                                 domain_name=domain_name,
                                 anti_affinity_ratio=1)


class TestClusterTemplate(BaseTestClusterTemplate):
    def _make_heat_template(self, cluster, ng1, ng2):
        heat_template = h.ClusterStack(cluster)
        heat_template.add_node_group_extra(ng1['id'], 1,
                                           get_ud_generator('line1\nline2'))
        heat_template.add_node_group_extra(ng2['id'], 1,
                                           get_ud_generator('line2\nline3'))
        return heat_template

    def test_get_anti_affinity_scheduler_hints(self):
        ng1, ng2 = self._make_node_groups('floating')
        cluster = self._make_cluster('private_net', ng1, ng2,
                                     anti_affinity=["datanode"])
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        ng1 = [ng for ng in cluster.node_groups if ng.name == "master"][0]
        ng2 = [ng for ng in cluster.node_groups if ng.name == "worker"][0]

        expected = {
            "scheduler_hints": {
                "group": {
                    "get_param": [h.SERVER_GROUP_NAMES, {"get_param":
                                                         "instance_index"}]
                }
            }
        }

        actual = heat_template._get_anti_affinity_scheduler_hints(ng2)
        self.assertEqual(expected, actual)

        expected = {}
        actual = heat_template._get_anti_affinity_scheduler_hints(ng1)
        self.assertEqual(expected, actual)

    def test_get_security_groups(self):
        ng1, ng2 = self._make_node_groups('floating')
        ng1['security_groups'] = ['1', '2']
        ng1['auto_security_group'] = False
        ng2['security_groups'] = ['3', '4']
        ng2['auto_security_group'] = True
        cluster = self._make_cluster('private_net', ng1, ng2)
        heat_template = self._make_heat_template(cluster, ng1, ng2)

        ng1 = [ng for ng in cluster.node_groups if ng.name == "master"][0]
        ng2 = [ng for ng in cluster.node_groups if ng.name == "worker"][0]
        expected = ['1', '2']
        actual = heat_template._get_security_groups(ng1)
        self.assertEqual(expected, actual)

        expected = ['3', '4', {'get_param': 'autosecgroup'}]
        actual = heat_template._get_security_groups(ng2)
        self.assertEqual(expected, actual)

    def test_get_security_groups_empty(self):
        ng1, _ = self._make_node_groups()
        ng1['security_groups'] = None
        ng1['auto_security_group'] = False
        cluster = self._make_cluster('private_net', ng1, ng1)
        heat_template = self._make_heat_template(cluster, ng1, ng1)

        ng1 = [ng for ng in cluster.node_groups if ng.name == "master"][0]
        actual = heat_template._get_security_groups(ng1)
        self.assertEqual([], actual)

    def _generate_auto_security_group_template(self):
        ng1, ng2 = self._make_node_groups('floating')
        cluster = self._make_cluster('private_net', ng1, ng2)
        ng1['cluster'] = cluster
        ng2['cluster'] = cluster
        ng1 = r.NodeGroupResource(ng1)
        ng2 = r.NodeGroupResource(ng2)
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        return heat_template._serialize_auto_security_group(ng1)

    @mock.patch('sahara.utils.openstack.neutron.get_private_network_cidrs')
    def test_serialize_auto_security_group_neutron(self, patched):
        ipv4_cidr = '192.168.0.0/24'
        ipv6_cidr = 'fe80::/64'
        patched.side_effect = lambda cluster: [ipv4_cidr, ipv6_cidr]
        expected_rules = [
            ('0.0.0.0/0', 'IPv4', 'tcp', '22', '22'),
            ('::/0', 'IPv6', 'tcp', '22', '22'),
            (ipv4_cidr, 'IPv4', 'tcp', '1', '65535'),
            (ipv4_cidr, 'IPv4', 'udp', '1', '65535'),
            (ipv4_cidr, 'IPv4', 'icmp', '0', '255'),
            (ipv6_cidr, 'IPv6', 'tcp', '1', '65535'),
            (ipv6_cidr, 'IPv6', 'udp', '1', '65535'),
            (ipv6_cidr, 'IPv6', 'icmp', '0', '255'),
        ]
        expected = {'cluster-master-1': {
            'type': 'OS::Neutron::SecurityGroup',
            'properties': {
                'description': 'Data Processing Cluster by Sahara\n'
                               'Sahara cluster name: cluster\n'
                               'Sahara engine: heat.3.0\n'
                               'Auto security group for Sahara Node '
                               'Group: master',
                'rules': [{
                    'remote_ip_prefix': rule[0],
                    'ethertype': rule[1],
                    'protocol': rule[2],
                    'port_range_min': rule[3],
                    'port_range_max': rule[4]
                } for rule in expected_rules]
            }
        }}
        actual = self._generate_auto_security_group_template()
        self.assertEqual(expected, actual)

    @mock.patch("sahara.conductor.objects.Cluster.use_designate_feature")
    def test_serialize_designate_records(self, mock_use_designate):
        ng1, ng2 = self._make_node_groups('floating')
        cluster = self._make_cluster('private_net', ng1, ng2,
                                     domain_name='domain.org.')

        mock_use_designate.return_value = False
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        expected = {}
        actual = heat_template._serialize_designate_records()
        self.assertEqual(expected, actual)

        mock_use_designate.return_value = True
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        expected = {
            'internal_designate_record': {
                'properties': {
                    'domain': 'domain.org.',
                    'name': {
                        'list_join': [
                            '.',
                            [{'get_attr': ['inst', 'name']}, 'domain.org.']]
                    },
                    'data': {'get_attr': ['inst', 'networks', 'private', 0]},
                    'type': 'A'
                },
                'type': 'OS::Designate::Record'
            },
            'external_designate_record': {
                'properties': {
                    'domain': 'domain.org.',
                    'name': {
                        'list_join': [
                            '.',
                            [{'get_attr': ['inst', 'name']}, 'domain.org.']]
                    },
                    'data': {'get_attr': ['floating_ip', 'ip']},
                    'type': 'A'
                },
                'type': 'OS::Designate::Record'
            }
        }
        actual = heat_template._serialize_designate_records()
        self.assertEqual(expected, actual)

    @mock.patch("sahara.conductor.objects.Cluster.use_designate_feature")
    def test_serialize_designate_reversed_records(self, mock_use_designate):

        def _generate_reversed_ip(ip):
            return {
                'list_join': [
                    '.',
                    [
                        {'str_split': ['.', ip, 3]},
                        {'str_split': ['.', ip, 2]},
                        {'str_split': ['.', ip, 1]},
                        {'str_split': ['.', ip, 0]},
                        'in-addr.arpa.'
                    ]
                ]
            }

        ng1, ng2 = self._make_node_groups('floating')
        cluster = self._make_cluster('private_net', ng1, ng2,
                                     domain_name='domain.org.')

        mock_use_designate.return_value = False
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        expected = {}
        actual = heat_template._serialize_designate_reverse_records()
        self.assertEqual(expected, actual)

        mock_use_designate.return_value = True
        heat_template = self._make_heat_template(cluster, ng1, ng2)
        expected = {
            'internal_designate_reverse_record': {
                'properties': {
                    'domain': 'in-addr.arpa.',
                    'name': _generate_reversed_ip(
                        {'get_attr': ['inst', 'networks', 'private', 0]}),
                    'data': {
                        'list_join': [
                            '.',
                            [{'get_attr': ['inst', 'name']}, 'domain.org.']]
                    },
                    'type': 'PTR'
                },
                'type': 'OS::Designate::Record'
            },
            'external_designate_reverse_record': {
                'properties': {
                    'domain': 'in-addr.arpa.',
                    'name': _generate_reversed_ip(
                        {'get_attr': ['floating_ip', 'ip']}),
                    'data': {
                        'list_join': [
                            '.',
                            [{'get_attr': ['inst', 'name']}, 'domain.org.']]
                    },
                    'type': 'PTR'
                },
                'type': 'OS::Designate::Record'
            }
        }
        actual = heat_template._serialize_designate_reverse_records()
        self.assertEqual(expected, actual)


class TestClusterTemplateWaitCondition(BaseTestClusterTemplate):
    def _make_heat_template(self, cluster, ng1, ng2):
        heat_template = h.ClusterStack(cluster)
        heat_template.add_node_group_extra(ng1.id, 1,
                                           get_ud_generator('line1\nline2'))
        heat_template.add_node_group_extra(ng2.id, 1,
                                           get_ud_generator('line2\nline3'))
        return heat_template

    def setUp(self):
        super(TestClusterTemplateWaitCondition, self).setUp()
        _ng1, _ng2 = self._make_node_groups("floating")
        _cluster = self._make_cluster("private_net", _ng1, _ng2)
        _ng1["cluster"] = _ng2["cluster"] = _cluster
        self.ng1 = mock.Mock()
        self.ng1.configure_mock(**_ng1)
        self.ng2 = mock.Mock()
        self.ng2.configure_mock(**_ng2)
        self.cluster = mock.Mock()
        self.cluster.configure_mock(**_cluster)
        self.template = self._make_heat_template(self.cluster,
                                                 self.ng1, self.ng2)

    @mock.patch('sahara.utils.cluster.etc_hosts_entry_for_service')
    def test_use_wait_condition(self, etc_hosts):
        etc_hosts.return_value = "data"
        self.override_config('heat_enable_wait_condition', True)
        instance = self.template._serialize_instance(self.ng1)
        expected_wc_handle = {
            "type": "OS::Heat::WaitConditionHandle"
        }
        expected_wc_waiter = {
            "type": "OS::Heat::WaitCondition",
            "depends_on": "inst",
            "properties": {
                "timeout": 3600,
                "handle": {"get_resource": "master-wc-handle"}
            }
        }
        self.assertEqual(expected_wc_handle, instance["master-wc-handle"])
        self.assertEqual(expected_wc_waiter, instance["master-wc-waiter"])


def get_ud_generator(s):
    def generator(*args, **kwargs):
        return s
    return generator



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xml.dom.minidom as xml

import pkg_resources as pkg
import testtools

from sahara.utils import xmlutils as x
from sahara import version


class XMLUtilsTestCase(testtools.TestCase):

    def setUp(self):
        super(XMLUtilsTestCase, self).setUp()

    def test_load_xml_defaults(self):
        self.assertEqual(
            [{'name': 'name1', 'value': 'value1', 'description': 'descr1'},
             {'name': 'name2', 'value': 'value2', 'description': 'descr2'},
             {'name': 'name3', 'value': '', 'description': 'descr3'},
             {'name': 'name4', 'value': '', 'description': 'descr4'},
             {'name': 'name5', 'value': 'value5', 'description': ''}],
            x.load_hadoop_xml_defaults(
                'tests/unit/resources/test-default.xml'))

    def test_parse_xml_with_name_and_value(self):
        file_name = 'tests/unit/resources/test-default.xml'
        fname = pkg.resource_filename(
            version.version_info.package, file_name)
        with open(fname, "r") as f:
            doc = "".join(line.strip() for line in f)
        self.assertEqual(
            [{'name': 'name1', 'value': 'value1'},
             {'name': 'name2', 'value': 'value2'},
             {'name': 'name3', 'value': ''},
             {'name': 'name4', 'value': ''},
             {'name': 'name5', 'value': 'value5'}],
            x.parse_hadoop_xml_with_name_and_value(doc)
        )

    def test_adjust_description(self):
        self.assertEqual("", x._adjust_field("\n"))
        self.assertEqual("", x._adjust_field("\n  "))
        self.assertEqual("abcdef", x._adjust_field("abc\n  def\n  "))
        self.assertEqual("abc de f", x._adjust_field("abc d\n e f\n"))
        self.assertEqual("abc", x._adjust_field("a\tb\t\nc"))

    def test_create_hadoop_xml(self):
        conf = x.load_hadoop_xml_defaults(
            'tests/unit/resources/test-default.xml')
        self.assertEqual("""<?xml version="1.0" ?>
<?xml-stylesheet type="text/xsl" href="configuration.xsl"?>
<configuration>
  <property>
    <name>name1</name>
    <value>some_val1</value>
  </property>
  <property>
    <name>name2</name>
    <value>2</value>
  </property>
</configuration>
""",
                         x.create_hadoop_xml({'name1': 'some_val1',
                                              'name2': 2}, conf),)

    def test_add_property_to_configuration(self):
        doc = self.create_default_doc()
        x.add_properties_to_configuration(doc, 'test', {'': 'empty1',
                                                        None: 'empty2'})
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <configuration/>
</test>
""",
                         doc.toprettyxml(indent="  "))

        test_conf = {'name1': 'value1', 'name2': 'value2'}
        x.add_properties_to_configuration(doc, 'test', test_conf)

        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <configuration>
    <property>
      <name>name1</name>
      <value>value1</value>
    </property>
    <property>
      <name>name2</name>
      <value>value2</value>
    </property>
  </configuration>
</test>
""",
                         doc.toprettyxml(indent="  "))
        x.add_property_to_configuration(doc, 'name3', 'value3')
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <configuration>
    <property>
      <name>name1</name>
      <value>value1</value>
    </property>
    <property>
      <name>name2</name>
      <value>value2</value>
    </property>
    <property>
      <name>name3</name>
      <value>value3</value>
    </property>
  </configuration>
</test>
""",
                         doc.toprettyxml(indent="  "))

    def test_get_if_not_exist_and_add_text_element(self):
        doc = self.create_default_doc()
        x.get_and_create_if_not_exist(doc, 'test', 'tag_to_add')
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <tag_to_add/>
</test>
""",
                         doc.toprettyxml(indent="  "))
        x.add_text_element_to_tag(doc, 'tag_to_add', 'p', 'v')
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <tag_to_add>
    <p>v</p>
  </tag_to_add>
</test>
""",
                         doc.toprettyxml(indent="  "))

    def test_get_if_not_exist_and_add_to_element(self):
        doc = self.create_default_doc()
        elem = x.get_and_create_if_not_exist(doc, 'test', 'tag_to_add')

        x.add_text_element_to_element(doc, elem, 'p', 'v')
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <tag_to_add>
    <p>v</p>
  </tag_to_add>
</test>
""",
                         doc.toprettyxml(indent="  "))

    def test_add_tagged_list(self):
        doc = self.create_default_doc()
        x.add_tagged_list(doc, 'test', 'list_item', ['a', 'b'])
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <list_item>a</list_item>
  <list_item>b</list_item>
</test>
""",
                         doc.toprettyxml(indent="  "))

    def test_add_equal_separated_dict(self):
        doc = self.create_default_doc()

        x.add_equal_separated_dict(doc, 'test', 'dict_item',
                                   {'': 'empty1', None: 'empty2'})
        self.assertEqual("""<?xml version="1.0" ?>
<test/>
""",
                         doc.toprettyxml(indent="  "))

        x.add_equal_separated_dict(doc, 'test', 'dict_item',
                                   {'a': 'b', 'c': 'd'})
        self.assertEqual("""<?xml version="1.0" ?>
<test>
  <dict_item>a=b</dict_item>
  <dict_item>c=d</dict_item>
</test>
""",
                         doc.toprettyxml(indent="  "))

    def create_default_doc(self):
        doc = xml.Document()
        test = doc.createElement('test')
        doc.appendChild(test)
        return doc

    def _get_xml_text(self, strip):
        doc = x.load_xml_document("service/edp/resources/workflow.xml", strip)
        x.add_child(doc, 'action', 'java')
        x.add_text_element_to_tag(doc, 'java', 'sometag', 'somevalue')
        return doc.toprettyxml(indent="  ").split("\n")

    def test_load_xml_document_strip(self):
        # Get the lines from the xml docs
        stripped = set(self._get_xml_text(True))
        unstripped = set(self._get_xml_text(False))

        # Prove they're different
        diff = stripped.symmetric_difference(unstripped)
        self.assertGreater(len(diff), 0)

        # Prove the differences are only blank lines
        non_blank_diffs = [l for l in diff if not l.isspace()]
        self.assertEqual(0, len(non_blank_diffs))



# Copyright (c) 2014 OpenStack Foundation
# Copyright (c) 2015 ISPRAS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock


import sahara.exceptions as ex
from sahara.service.edp.job_utils import ds_manager
from sahara.service.edp.spark import engine as se
from sahara.tests.unit import base
from sahara.utils import edp


class TestSpark(base.SaharaTestCase):
    def setUp(self):
        super(TestSpark, self).setUp()

        # These variables are initialized in subclasses because its values
        # depend on plugin
        self.master_host = None
        self.engine_class = None
        self.spark_user = None
        self.spark_submit = None
        self.master = None
        self.deploy_mode = None

        self.master_port = 7077
        self.master_inst = "6789"
        self.spark_pid = "12345"
        self.spark_home = "/opt/spark"
        self.workflow_dir = "/wfdir"
        self.driver_cp = "/usr/lib/hadoop-mapreduce/hadoop-openstack.jar:"

        ds_manager.setup_data_sources()

    def test_get_pid_and_inst_id(self):
        '''Test parsing of job ids

        Test that job ids of the form pid@instance are
        split into pid and instance ids by eng._get_pid_and_inst_id()
        but anything else returns empty strings
        '''
        eng = se.SparkJobEngine(None)
        for job_id in [None, "", "@", "something", "pid@", "@instance"]:
            pid, inst_id = eng._get_pid_and_inst_id(job_id)
            self.assertEqual(("", ""), (pid, inst_id))

        pid, inst_id = eng._get_pid_and_inst_id("pid@instance")
        self.assertEqual(("pid", "instance"), (pid, inst_id))

    @mock.patch('sahara.utils.cluster.get_instances')
    def test_get_instance_if_running(self, get_instances):
        '''Test retrieval of pid and instance object for running job

        If the job id is valid and the job status is non-terminated,
        _get_instance_if_running() should retrieve the instance
        based on the inst_id and return the pid and instance.

        If the job is invalid or the job is terminated, it should
        return None, None.

        If get_instances() throws an exception or returns an empty list,
        the instance returned should be None (pid might still be set)
        '''
        get_instances.return_value = ["instance"]
        job_exec = mock.Mock()
        eng = se.SparkJobEngine("cluster")

        job_exec.engine_job_id = "invalid id"
        self.assertEqual((None, None),
                         eng._get_instance_if_running(job_exec))

        job_exec.engine_job_id = "pid@inst_id"
        for state in edp.JOB_STATUSES_TERMINATED:
            job_exec.info = {'status': state}
            self.assertEqual((None, None),
                             eng._get_instance_if_running(job_exec))

        job_exec.info = {'status': edp.JOB_STATUS_RUNNING}
        self.assertEqual(("pid", "instance"),
                         eng._get_instance_if_running(job_exec))
        get_instances.assert_called_with("cluster", ["inst_id"])

        # Pretend get_instances returns nothing
        get_instances.return_value = []
        pid, instance = eng._get_instance_if_running(job_exec)
        self.assertIsNone(instance)

        # Pretend get_instances throws an exception
        get_instances.side_effect = Exception("some failure")
        pid, instance = eng._get_instance_if_running(job_exec)
        self.assertIsNone(instance)

    def test_get_result_file(self):
        remote = mock.Mock()
        remote.execute_command.return_value = 999, "value"
        job_exec = mock.Mock()
        job_exec.extra = {"spark-path": "/tmp/spark-edp/Job/123"}

        eng = se.SparkJobEngine("cluster")
        ret, stdout = eng._get_result_file(remote, job_exec)
        remote.execute_command.assert_called_with(
            "cat /tmp/spark-edp/Job/123/result",
            raise_when_error=False)
        self.assertEqual((ret, stdout),
                         remote.execute_command.return_value)

    def test_check_pid(self):
        remote = mock.Mock()
        remote.execute_command.return_value = 999, ""

        eng = se.SparkJobEngine("cluster")
        ret = eng._check_pid(remote, "pid")
        remote.execute_command.assert_called_with("ps hp pid",
                                                  raise_when_error=False)
        self.assertEqual(999, ret)

    @mock.patch.object(se.SparkJobEngine,
                       '_get_result_file',
                       autospec=True)
    @mock.patch.object(se.SparkJobEngine,
                       '_check_pid',
                       autospec=True)
    def test_get_job_status_from_remote(self, _check_pid, _get_result_file):
        '''Test retrieval of job status from remote instance

        If the process is present, status is RUNNING
        If the process is not present, status depends on the result file
        If the result file is missing, status is DONEWITHERROR
        '''
        eng = se.SparkJobEngine("cluster")
        job_exec = mock.Mock()
        remote = mock.Mock()

        # Pretend process is running
        _check_pid.return_value = 0
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        _check_pid.assert_called_with(eng, remote, "pid")
        self.assertEqual({"status": edp.JOB_STATUS_RUNNING}, status)

        # Pretend process ended and result file contains 0 (success)
        _check_pid.return_value = 1
        _get_result_file.return_value = 0, "0"
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_SUCCEEDED}, status)

        # Pretend process ended and result file contains 1 (success)
        _get_result_file.return_value = 0, "1"
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_DONEWITHERROR}, status)

        # Pretend process ended and result file contains 130 (killed)
        _get_result_file.return_value = 0, "130"
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_KILLED}, status)

        # Pretend process ended and result file contains -2 (killed)
        _get_result_file.return_value = 0, "-2"
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_KILLED}, status)

        # Pretend process ended and result file is missing
        _get_result_file.return_value = 1, ""
        status = eng._get_job_status_from_remote(remote, "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_DONEWITHERROR}, status)

    @mock.patch.object(se.SparkJobEngine,
                       '_get_job_status_from_remote',
                       autospec=True)
    @mock.patch.object(se.SparkJobEngine,
                       '_get_instance_if_running',
                       autospec=True)
    @mock.patch('sahara.utils.remote.get_remote')
    def test_get_job_status(self,
                            get_remote,
                            _get_instance_if_running,
                            _get_job_status_from_remote):

        # This is to mock "with remote.get_remote(instance) as r"
        remote_instance = mock.Mock()
        get_remote.return_value.__enter__ = mock.Mock(
            return_value=remote_instance)

        # Pretend instance is not returned
        _get_instance_if_running.return_value = "pid", None
        job_exec = mock.Mock()
        eng = se.SparkJobEngine("cluster")
        status = eng.get_job_status(job_exec)
        self.assertIsNone(status)

        # Pretend we have an instance
        _get_instance_if_running.return_value = "pid", "instance"
        _get_job_status_from_remote.return_value = {"status":
                                                    edp.JOB_STATUS_RUNNING}
        status = eng.get_job_status(job_exec)
        _get_job_status_from_remote.assert_called_with(eng,
                                                       remote_instance,
                                                       "pid", job_exec)
        self.assertEqual({"status": edp.JOB_STATUS_RUNNING}, status)

    @mock.patch.object(se.SparkJobEngine,
                       '_get_instance_if_running',
                       autospec=True,
                       return_value=(None, None))
    @mock.patch('sahara.utils.remote.get_remote')
    def test_cancel_job_null_or_done(self,
                                     get_remote,
                                     _get_instance_if_running):
        '''Test cancel_job() when instance is None

        Test that cancel_job() returns None and does not try to
        retrieve a remote instance if _get_instance_if_running() returns None
        '''
        eng = se.SparkJobEngine("cluster")
        job_exec = mock.Mock()
        self.assertIsNone(eng.cancel_job(job_exec))
        self.assertTrue(_get_instance_if_running.called)
        self.assertFalse(get_remote.called)

    @mock.patch.object(se.SparkJobEngine,
                       '_get_job_status_from_remote',
                       autospec=True,
                       return_value={"status": edp.JOB_STATUS_KILLED})
    @mock.patch.object(se.SparkJobEngine,
                       '_get_instance_if_running',
                       autospec=True,
                       return_value=("pid", "instance"))
    @mock.patch('sahara.utils.remote.get_remote')
    def test_cancel_job(self,
                        get_remote,
                        _get_instance_if_running,
                        _get_job_status_from_remote):
        '''Test cancel_job() with a valid instance

        For a valid instance, test that cancel_job:

        * retrieves the remote instance
        * executes the proper kill command
        * retrieves the job status (because the remote command is successful)
        '''

        # This is to mock "with remote.get_remote(instance) as r" in cancel_job
        # and to mock r.execute_command to return success
        remote_instance = mock.Mock()
        get_remote.return_value.__enter__ = mock.Mock(
            return_value=remote_instance)
        remote_instance.execute_command.return_value = (0, "standard out")

        eng = se.SparkJobEngine("cluster")
        job_exec = mock.Mock()
        status = eng.cancel_job(job_exec)

        # check that remote.get_remote was called with the result of
        # eng._get_instance_if_running()
        get_remote.assert_called_with("instance")

        # check that execute_command was called with the proper arguments
        # ("pid" was passed in)
        remote_instance.execute_command.assert_called_with(
            "kill -SIGINT pid",
            raise_when_error=False)

        # check that the job status was retrieved since the command succeeded
        _get_job_status_from_remote.assert_called_with(eng,
                                                       remote_instance,
                                                       "pid", job_exec)

        self.assertEqual({"status": edp.JOB_STATUS_KILLED}, status)

    @mock.patch.object(se.SparkJobEngine,
                       '_get_job_status_from_remote',
                       autospec=True)
    @mock.patch.object(se.SparkJobEngine,
                       '_get_instance_if_running',
                       autospec=True,
                       return_value=("pid", "instance"))
    @mock.patch('sahara.utils.remote.get_remote')
    def test_cancel_job_failed(self,
                               get_remote,
                               _get_instance_if_running,
                               _get_job_status_from_remote):
        '''Test cancel_job() when remote command fails

        For a valid instance and a failed kill command, test that cancel_job:

        * retrieves the remote instance
        * executes the proper kill command
        * does not retrieve the job status (because the remote command failed)
        '''

        # This is to mock "with remote.get_remote(instance) as r"
        # and to mock r.execute_command to return failure
        remote_instance = mock.Mock()
        get_remote.return_value.__enter__ = mock.Mock(
            return_value=remote_instance)
        remote_instance.execute_command.return_value = (-1, "some error")

        eng = se.SparkJobEngine("cluster")
        job_exec = mock.Mock()
        status = eng.cancel_job(job_exec)

        # check that remote.get_remote was called with the result of
        # eng._get_instance_if_running
        get_remote.assert_called_with("instance")

        # check that execute_command was called with the proper arguments
        # ("pid" was passed in)
        remote_instance.execute_command.assert_called_with(
            "kill -SIGINT pid",
            raise_when_error=False)

        # check that the job status was not retrieved since the command failed
        self.assertEqual(0, _get_job_status_from_remote.called)

        # check that we have nothing new to report ...
        self.assertIsNone(status)

    @mock.patch('sahara.service.edp.spark.engine.jb_manager')
    @mock.patch('sahara.utils.remote.get_remote')
    def test_upload_job_files(self, get_remote, jb_manager):
        main_names = ["main1", "main2", "main3"]
        lib_names = ["lib1", "lib2", "lib3"]

        def make_data_objects(*args):
            objs = []
            for name in args:
                m = mock.Mock()
                m.name = name
                objs.append(m)
            return objs

        job = mock.Mock()
        job.name = "job"
        job.mains = make_data_objects(*main_names)
        job.libs = make_data_objects(*lib_names)

        # This is to mock "with remote.get_remote(instance) as r"
        remote_instance = mock.Mock()
        remote_instance.instance.node_group.cluster.shares = []
        remote_instance.instance.node_group.shares = []
        get_remote.return_value.__enter__ = mock.Mock(
            return_value=remote_instance)

        JOB_BINARIES = mock.Mock()
        mock_jb = mock.Mock()
        jb_manager.JOB_BINARIES = JOB_BINARIES

        JOB_BINARIES.get_job_binary_by_url = mock.Mock(return_value=mock_jb)

        mock_jb.copy_binary_to_cluster = mock.Mock(side_effect=[
                                                   '/somedir/main1',
                                                   '/somedir/main2',
                                                   '/somedir/main3',
                                                   '/somedir/lib1',
                                                   '/somedir/lib2',
                                                   '/somedir/lib3'])

        eng = se.SparkJobEngine("cluster")
        eng._prepare_job_binaries = mock.Mock()

        paths, builtins = eng._upload_job_files("where", "/somedir", job, {})
        self.assertEqual(["/somedir/" + n for n in main_names + lib_names],
                         paths)

    def _make_master_instance(self, return_code=0):
        master = mock.Mock()
        master.execute_command.return_value = (return_code, self.spark_pid)
        master.get_python_version.return_value = 'python'
        master.hostname.return_value = self.master_host
        master.id = self.master_inst
        return master

    def _config_values(self, *key):
        return {("Spark", "Master port", "cluster"): self.master_port,
                ("Spark", "Spark home", "cluster"): self.spark_home,
                ("Spark", "Executor extra classpath",
                 "cluster"): self.driver_cp}[key]

    @mock.patch('sahara.conductor.API.job_execution_update')
    @mock.patch('sahara.conductor.API.job_execution_get')
    @mock.patch('sahara.utils.remote.get_remote')
    @mock.patch('sahara.plugins.utils.get_config_value_or_default')
    @mock.patch('sahara.service.edp.job_utils.create_workflow_dir')
    @mock.patch('sahara.plugins.utils.get_instance')
    @mock.patch('sahara.conductor.API.job_get')
    @mock.patch('sahara.context.ctx', return_value="ctx")
    def _setup_run_job(self, master_instance, job_configs, files,
                       ctx, job_get, get_instance, create_workflow_dir,
                       get_config_value, get_remote, job_exec_get,
                       job_exec_update):

        def _upload_job_files(where, job_dir, job,
                              libs_subdir=True, job_configs=None):
            paths = [os.path.join(self.workflow_dir, f) for f in files['jars']]
            bltns = files.get('bltns', [])
            bltns = [os.path.join(self.workflow_dir, f) for f in bltns]
            return paths, bltns

        job = mock.Mock()
        job.name = "MyJob"
        job_get.return_value = job

        job_exec = mock.Mock()
        job_exec.job_configs = job_configs

        get_config_value.side_effect = self._config_values

        create_workflow_dir.return_value = self.workflow_dir

        # This is to mock "with remote.get_remote(master) as r" in run_job
        get_remote.return_value.__enter__ = mock.Mock(
            return_value=master_instance)
        get_instance.return_value = master_instance

        eng = self.engine_class("cluster")
        eng._upload_job_files = mock.Mock()
        eng._upload_job_files.side_effect = _upload_job_files
        status = eng.run_job(job_exec)

        # Check that we launch on the master node
        get_instance.assert_called_with("cluster", self.master_host)

        return status

    def test_run_job_raise(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass"},
            'args': ['input_arg', 'output_arg']
        }

        files = {'jars': ["app.jar",
                          "jar1.jar",
                          "jar2.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance(return_code=1)

        # If execute_command returns an error we should get a raise
        self.assertRaises(ex.EDPError,
                          self._setup_run_job,
                          master_instance, job_configs, files)

    def test_run_job_extra_jars_args(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass"},
            'args': ['input_arg', 'output_arg']
        }

        files = {'jars': ["app.jar",
                          "jar1.jar",
                          "jar2.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--class org.me.myclass --jars jar1.jar,jar2.jar '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar input_arg output_arg '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    def test_run_job_args(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass"},
            'args': ['input_arg', 'output_arg']
        }

        files = {'jars': ["app.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--class org.me.myclass '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar input_arg output_arg '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    def test_run_job(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass"},
        }

        files = {'jars': ["app.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--class org.me.myclass '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    def test_run_job_wrapper_extra_jars_args(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass",
                        "edp.spark.adapt_for_swift": True},
            'args': ['input_arg', 'output_arg']
        }

        files = {'jars': ["app.jar",
                          "jar1.jar",
                          "jar2.jar"],
                 'bltns': ["wrapper.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--driver-class-path %(driver_cp)s '
            '--files spark.xml '
            '--class org.openstack.sahara.edp.SparkWrapper '
            '--jars wrapper.jar,jar1.jar,jar2.jar '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar spark.xml org.me.myclass input_arg output_arg '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "driver_cp": self.driver_cp,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    def test_run_job_wrapper_args(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass",
                        "edp.spark.adapt_for_swift": True},
            'args': ['input_arg', 'output_arg']
        }

        files = {'jars': ["app.jar"],
                 'bltns': ["wrapper.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--driver-class-path %(driver_cp)s '
            '--files spark.xml '
            '--class org.openstack.sahara.edp.SparkWrapper '
            '--jars wrapper.jar '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar spark.xml org.me.myclass input_arg output_arg '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "driver_cp": self.driver_cp,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    def test_run_job_wrapper(self):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass",
                        "edp.spark.adapt_for_swift": True}
        }

        files = {'jars': ["app.jar"],
                 'bltns': ["wrapper.jar"]}

        # The object representing the spark master node
        # The spark-submit command will be run on this instance
        master_instance = self._make_master_instance()
        status = self._setup_run_job(master_instance, job_configs, files)

        # Check the command
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--driver-class-path %(driver_cp)s '
            '--files spark.xml '
            '--class org.openstack.sahara.edp.SparkWrapper '
            '--jars wrapper.jar '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar spark.xml org.me.myclass '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "driver_cp": self.driver_cp,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})

        # Check result here
        self.assertEqual(("%s@%s" % (self.spark_pid, self.master_inst),
                          edp.JOB_STATUS_RUNNING,
                          {"spark-path": self.workflow_dir}), status)

    @mock.patch('sahara.service.edp.job_utils.prepare_cluster_for_ds')
    @mock.patch('sahara.service.edp.job_utils.resolve_data_source_references')
    def test_external_hdfs_config(self, resolver, prepare):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass"},
        }

        files = {'jars': ["app.jar"]}

        data_source = mock.Mock()
        data_source.type = 'hdfs'
        data_source.id = 'id'
        resolver.return_value = ([data_source], job_configs)

        master_instance = self._make_master_instance()
        self._setup_run_job(master_instance, job_configs, files)

        prepare.assert_called_once()

    @mock.patch('sahara.service.edp.job_utils.prepare_cluster_for_ds')
    @mock.patch('sahara.service.edp.job_utils.resolve_data_source_references')
    def test_overridden_driver_classpath(self, resolver, prepare):
        job_configs = {
            'configs': {"edp.java.main_class": "org.me.myclass",
                        'edp.spark.driver.classpath': "my-classpath.jar"},
        }

        files = {'jars': ["app.jar"]}

        data_source = mock.Mock()
        data_source.type = 'hdfs'
        data_source.id = 'id'
        resolver.return_value = ([data_source], job_configs)

        master_instance = self._make_master_instance()
        self._setup_run_job(master_instance, job_configs, files)

        # check that overridden value was applied
        master_instance.execute_command.assert_called_with(
            'cd %(workflow_dir)s; '
            './launch_command %(spark_user)s%(spark_submit)s '
            '--driver-class-path my-classpath.jar '
            '--class org.me.myclass '
            '--master %(master)s '
            '--deploy-mode %(deploy_mode)s '
            'app.jar '
            '> /dev/null 2>&1 & echo $!' % {"workflow_dir": self.workflow_dir,
                                            "spark_user": self.spark_user,
                                            "spark_submit": self.spark_submit,
                                            "master": self.master,
                                            "deploy_mode": self.deploy_mode})



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from oslo_config import cfg
from oslo_log import log
import six

from sahara.cli.image_pack import api
from sahara.i18n import _

LOG = log.getLogger(__name__)

CONF = cfg.CONF

CONF.register_cli_opts([
    cfg.StrOpt(
        'image',
        required=True,
        help=_("The path to an image to modify. This image will be modified "
               "in-place: be sure to target a copy if you wish to maintain a "
               "clean master image.")),
    cfg.StrOpt(
        'root-filesystem',
        dest='root_fs',
        required=False,
        help=_("The filesystem to mount as the root volume on the image. No "
               "value is required if only one filesystem is detected.")),
    cfg.BoolOpt(
        'test-only',
        dest='test_only',
        default=False,
        help=_("If this flag is set, no changes will be made to the image; "
               "instead, the script will fail if discrepancies are found "
               "between the image and the intended state."))])


def unregister_extra_cli_opt(name):
    try:
        for cli in CONF._cli_opts:
            if cli['opt'].name == name:
                CONF.unregister_opt(cli['opt'])
    except Exception:
        pass


for extra_opt in ["log-exchange", "host", "port"]:
    unregister_extra_cli_opt(extra_opt)


def add_plugin_parsers(subparsers):
    api.setup_plugins()
    for plugin in api.get_loaded_plugins():
        args_by_version = api.get_plugin_arguments(plugin)
        if all(args is NotImplemented for version, args
               in six.iteritems(args_by_version)):
            continue
        plugin_parser = subparsers.add_parser(
            plugin, help=_('Image generation for the {plugin} plugin').format(
                plugin=plugin))
        version_parsers = plugin_parser.add_subparsers(
            title=_("Plugin version"),
            dest="version",
            help=_("Available versions"))
        for version, args in six.iteritems(args_by_version):
            if args is NotImplemented:
                continue
            version_parser = version_parsers.add_parser(
                version, help=_('{plugin} version {version}').format(
                    plugin=plugin, version=version))
            for arg in args:
                arg_token = ("--%s" % arg.name if len(arg.name) > 1 else
                             "-%s" % arg.name)
                version_parser.add_argument(arg_token,
                                            dest=arg.name,
                                            help=arg.description,
                                            default=arg.default,
                                            required=arg.required,
                                            choices=arg.choices)
            version_parser.set_defaults(args={arg.name
                                              for arg in args})


command_opt = cfg.SubCommandOpt('plugin',
                                title=_('Plugin'),
                                help=_('Available plugins'),
                                handler=add_plugin_parsers)

CONF.register_cli_opt(command_opt)


def main():
    CONF(project='sahara')

    CONF.reload_config_files()
    log.setup(CONF, "sahara")

    LOG.info("Command: {command}".format(command=' '.join(sys.argv)))

    api.set_logger(LOG)
    api.set_conf(CONF)

    plugin = CONF.plugin.name
    version = CONF.plugin.version
    args = CONF.plugin.args
    image_arguments = {arg: getattr(CONF.plugin, arg) for arg in args}

    api.pack_image(CONF.image, plugin, version, image_arguments,
                   CONF.root_fs, CONF.test_only)

    LOG.info("Finished packing image for {plugin} at version "
             "{version}".format(plugin=plugin, version=version))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import threading

from oslo_config import cfg
from oslo_log import log as logging

from sahara import conductor as c
from sahara import context
from sahara.i18n import _
from sahara.plugins import provisioning as plugin_base
from sahara.utils import cluster_progress_ops as cpo
from sahara.utils.openstack import base as b
from sahara.utils.openstack import cinder
from sahara.utils.openstack import nova
from sahara.utils import poll_utils


conductor = c.API
LOG = logging.getLogger(__name__)

CONF = cfg.CONF


def _get_timeout_for_disk_preparing(cluster):
    configs = cluster.cluster_configs.to_dict()
    option_name = plugin_base.DISKS_PREPARING_TIMEOUT.name
    option_target = plugin_base.DISKS_PREPARING_TIMEOUT.applicable_target
    try:
        return int(configs[option_target][option_name])
    except Exception:
        return int(plugin_base.DISKS_PREPARING_TIMEOUT.default_value)


def _is_xfs_enabled(cluster):
    configs = cluster.cluster_configs.to_dict()
    option_name = plugin_base.XFS_ENABLED.name
    option_target = plugin_base.XFS_ENABLED.applicable_target
    try:
        return bool(configs[option_target][option_name])
    except Exception:
        return bool(plugin_base.XFS_ENABLED.default_value)


def _get_os_distrib(remote):
    return remote.get_os_distrib()


def _check_installed_xfs(instance):
    redhat = "rpm -q xfsprogs || yum install -y xfsprogs"
    debian = "dpkg -s xfsprogs || apt-get -y install xfsprogs"

    cmd_map = {
        "centos": redhat,
        "fedora": redhat,
        "redhatenterpriseserver": redhat,
        "redhat": redhat,
        "ubuntu": debian,
        'debian': debian
    }

    with instance.remote() as r:
        distro = _get_os_distrib(r)
        if not cmd_map.get(distro):
            LOG.warning("Cannot verify installation of XFS tools for "
                        "unknown distro {distro}.".format(distro=distro))
            return False
        try:
            r.execute_command(cmd_map.get(distro), run_as_root=True)
            return True
        except Exception as e:
            LOG.warning("Cannot install xfsprogs: {reason}".format(reason=e))
            return False


def _can_use_xfs(instances):
    cluster = instances[0].cluster
    if not _is_xfs_enabled(cluster):
        return False
    for instance in instances:
        if not _check_installed_xfs(instance):
            return False
    return True


def mount_to_instances(instances):
    if len(instances) == 0:
        return

    use_xfs = _can_use_xfs(instances)

    for instance in instances:
        with context.set_current_instance_id(instance.instance_id):
            devices = _find_instance_devices(instance)

            if devices:
                cpo.add_provisioning_step(
                    instance.cluster_id,
                    _("Mount volumes to {inst_name} instance").format(
                        inst_name=instance.instance_name), len(devices))

                formatted_devices = []
                lock = threading.Lock()
                with context.ThreadGroup() as tg:
                    # Since formating can take several minutes (for large
                    # disks) and can be done in parallel, launch one thread
                    # per disk.
                    for device in devices:
                        tg.spawn('format-device-%s' % device, _format_device,
                                 instance, device, use_xfs, formatted_devices,
                                 lock)

                conductor.instance_update(
                    context.current(), instance,
                    {"storage_devices_number": len(formatted_devices)})
                for idx, dev in enumerate(formatted_devices):
                    _mount_volume_to_node(instance, idx+1, dev, use_xfs)


def _find_instance_devices(instance):
    with instance.remote() as r:
        code, attached_info = r.execute_command(
            "lsblk -r | awk '$6 ~ /disk/ || /part/ {print \"/dev/\" $1}'")
        attached_dev = attached_info.split()
        code, mounted_info = r.execute_command(
            "mount | awk '$1 ~ /^\/dev/ {print $1}'")
        mounted_dev = mounted_info.split()

        # find and ignore Nova config drive
        for label in ("config-2", "CONFIG-2"):
            code, nova_config_drive = r.execute_command(
                "/sbin/blkid -t LABEL=\"%s\" -odevice" % label,
                raise_when_error=False,
                run_as_root=True
            )
            drive_name = nova_config_drive.strip()
            if code == 0 and drive_name in attached_dev:
                attached_dev.remove(drive_name)
                break

    # filtering attached devices, that should not be mounted
    for dev in attached_dev[:]:
        idx = re.sub("\D", "", dev)
        if idx:
            if dev in mounted_dev:
                if re.sub("\d", "", dev) in attached_dev:
                    attached_dev.remove(re.sub("\d", "", dev))
                attached_dev.remove(dev)

    for dev in attached_dev[:]:
        if re.sub("\D", "", dev):
            if re.sub("\d", "", dev) in attached_dev:
                attached_dev.remove(dev)

    attached_dev = [dev for dev in attached_dev if dev not in mounted_dev]

    return attached_dev


@cpo.event_wrapper(mark_successful_on_exit=True)
def _mount_volume_to_node(instance, index, device, use_xfs):
    LOG.debug("Mounting volume {device} to instance".format(device=device))
    mount_point = instance.node_group.volume_mount_prefix + str(index)
    _mount_volume(instance, device, mount_point, use_xfs)
    LOG.debug("Mounted volume to instance")


def _format_device(
        instance, device, use_xfs, formatted_devices=None, lock=None):
    with instance.remote() as r:
        try:
            timeout = _get_timeout_for_disk_preparing(instance.cluster)

            # Format devices with better performance options:
            # - reduce number of blocks reserved for root to 1%
            # - use 'dir_index' for faster directory listings
            # - use 'extents' to work faster with large files
            # - disable journaling
            fs_opts = '-F -m 1 -O dir_index,extents,^has_journal'
            command = 'sudo mkfs.ext4 %s %s' % (fs_opts, device)
            if use_xfs:
                command = 'sudo mkfs.xfs -f %s' % device
            r.execute_command(command, timeout=timeout)
            if lock:
                with lock:
                    formatted_devices.append(device)
        except Exception as e:
            LOG.warning("Device {dev} cannot be formatted: {reason}".format(
                        dev=device, reason=e))
            cpo.add_fail_event(instance, e)


def _mount_volume(instance, device_path, mount_point, use_xfs):
    with instance.remote() as r:
        try:
            timeout = _get_timeout_for_disk_preparing(instance.cluster)

            # Mount volumes with better performance options:
            # - enable write-back for ext4
            # - do not store access time
            # - disable barrier for xfs

            r.execute_command('sudo mkdir -p %s' % mount_point)
            mount_opts = '-o data=writeback,noatime,nodiratime'
            if use_xfs:
                mount_opts = "-t xfs -o noatime,nodiratime,nobarrier"

            r.execute_command('sudo mount %s %s %s' %
                              (mount_opts, device_path, mount_point),
                              timeout=timeout)
            r.execute_command(
                'sudo sh -c "grep %s /etc/mtab >> /etc/fstab"' % device_path)

        except Exception:
            LOG.error("Error mounting volume to instance")
            raise


def detach_from_instance(instance):
    for volume_id in instance.volumes:
        _detach_volume(instance, volume_id)
        _delete_volume(volume_id)


@poll_utils.poll_status(
    'detach_volume_timeout', _("Await for volume become detached"), sleep=2)
def _await_detach(volume_id):
    volume = cinder.get_volume(volume_id)
    if volume.status not in ['available', 'error']:
        return False
    return True


def _detach_volume(instance, volume_id):
    volume = cinder.get_volume(volume_id)
    try:
        LOG.debug("Detaching volume {id} from instance".format(id=volume_id))
        b.execute_with_retries(nova.client().volumes.delete_server_volume,
                               instance.instance_id, volume_id)
    except Exception:
        LOG.error("Can't detach volume {id}".format(id=volume.id))

    detach_timeout = CONF.timeouts.detach_volume_timeout
    LOG.debug("Waiting {timeout} seconds to detach {id} volume".format(
              timeout=detach_timeout, id=volume_id))
    _await_detach(volume_id)


def _delete_volume(volume_id):
    LOG.debug("Deleting volume {volume}".format(volume=volume_id))
    volume = cinder.get_volume(volume_id)
    try:
        b.execute_with_retries(volume.delete)
    except Exception:
        LOG.error("Can't delete volume {volume}".format(volume=volume.id))



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.api import acl
from sahara.service.api.v2 import plugins as api
from sahara.service import validation as v
from sahara.service.validations import plugins as v_p
import sahara.utils.api as u


rest = u.RestV2('plugins', __name__)


@rest.get('/plugins')
@acl.enforce("data-processing:plugin:list")
@v.validate_request_params([])
def plugins_list():
    return u.render(plugins=[p.dict for p in api.get_plugins()])


@rest.get('/plugins/<plugin_name>')
@acl.enforce("data-processing:plugin:get")
@v.check_exists(api.get_plugin, plugin_name='plugin_name')
@v.validate_request_params([])
def plugins_get(plugin_name):
    return u.render(api.get_plugin(plugin_name).wrapped_dict)


@rest.get('/plugins/<plugin_name>/<version>')
@acl.enforce("data-processing:plugin:get-version")
@v.check_exists(api.get_plugin, plugin_name='plugin_name', version='version')
@v.validate_request_params([])
def plugins_get_version(plugin_name, version):
    return u.render(api.get_plugin(plugin_name, version).wrapped_dict)


@rest.patch('/plugins/<plugin_name>')
@acl.enforce("data-processing:plugin:update")
@v.check_exists(api.get_plugin, plugin_name='plugin_name')
@v.validate(v_p.plugin_update_validation_jsonschema(), v_p.check_plugin_update)
@v.validate_request_params([])
def plugins_update(plugin_name, data):
    return u.render(api.update_plugin(plugin_name, data).wrapped_dict)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from manilaclient.common.apiclient import exceptions as manila_ex
except ImportError:
    from manilaclient.openstack.common.apiclient import exceptions as manila_ex
from unittest import mock

from sahara.service.validations import shares
from sahara.tests.unit.service.validation import utils as u


class TestShareValidations(u.ValidationTestCase):

    def setUp(self):
        super(TestShareValidations, self).setUp()
        self._create_object_fun = shares.check_shares
        self.scheme = shares.SHARE_SCHEMA

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares(self, f_client):
        f_client.return_value = mock.Mock(
            shares=mock.Mock(
                get=mock.Mock(
                    return_value=mock.Mock(share_proto='NFS'))))

        self._assert_create_object_validation(data=[
            {
                "id": "12345678-1234-1234-1234-123456789012",
                "path": "/path",
                "access_level": 'rw'
            }])

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares_bad_type(self, f_client):
        f_client.return_value = mock.Mock(
            shares=mock.Mock(
                get=mock.Mock(
                    return_value=mock.Mock(share_proto='Mackerel'))))

        self._assert_create_object_validation(
            data=[
                {
                    "id": "12345678-1234-1234-1234-123456789012",
                    "path": "/path",
                    "access_level": 'rw'
                }],
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Requested share id "
                       "12345678-1234-1234-1234-123456789012 is of type "
                       "Mackerel, which is not supported by Sahara."))

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares_overlapping_paths(self, f_client):

        self._assert_create_object_validation(
            data=[
                {
                    "id": "12345678-1234-1234-1234-123456789012",
                    "path": "/path",
                },
                {
                    "id": "DEADBEEF-DEAD-BEEF-DEAD-BEEFDEADBEEF",
                    "path": "/path"
                }],
            bad_req_i=(1, 'INVALID_DATA',
                       "Multiple shares cannot be mounted to the same path."))

        self.assertEqual(0, f_client.call_count)

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares_no_share_exists(self, f_client):
        f_client.return_value = mock.Mock(
            shares=mock.Mock(
                get=mock.Mock(
                    side_effect=manila_ex.NotFound)))

        self._assert_create_object_validation(
            data=[
                {
                    "id": "12345678-1234-1234-1234-123456789012",
                    "path": "/path"
                }],
            bad_req_i=(
                1, 'INVALID_REFERENCE',
                "Requested share id 12345678-1234-1234-1234-123456789012 does "
                "not exist."))

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares_bad_paths(self, f_client):

        self._assert_create_object_validation(
            data=[
                {
                    "id": "12345678-1234-1234-1234-123456789012",
                    "path": "path"
                }],
            bad_req_i=(
                1, 'INVALID_DATA',
                'Paths must be absolute Linux paths starting with "/" '
                'and may not contain nulls.'))

        self._assert_create_object_validation(
            data=[
                {
                    "id": "12345678-1234-1234-1234-123456789012",
                    "path": "\x00"
                }],
            bad_req_i=(
                1, 'INVALID_DATA',
                'Paths must be absolute Linux paths starting with "/" '
                'and may not contain nulls.'))

        self.assertEqual(0, f_client.call_count)

    @mock.patch('sahara.utils.openstack.manila.client')
    def test_shares_no_shares(self, f_client):
        self._assert_create_object_validation(data=[])
        self.assertEqual(0, f_client.call_count)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections
import itertools

from oslo_log import log
import six

from sahara import context
from sahara.utils.openstack import manila

LOG = log.getLogger(__name__)


def mount_shares(cluster):
    """Mounts all shares specified for the cluster and any of its node groups.

    - In the event that a specific share is configured for both the cluster and
      a specific node group, configuration at the node group level will be
      ignored.
    - In the event that utilities required to mount the share are not
      already installed on the node, this method may fail if the node cannot
      access the internet.
    - This method will not remove already-mounted shares.
    - This method will not remove or remount (or currently, reconfigure) shares
      already mounted to the desired local mount point.

    :param cluster: The cluster model.
    """

    node_groups = (ng for ng in cluster.node_groups if ng.shares)
    ng_mounts = [_mount(ng, share_config) for ng in node_groups
                 for share_config in ng.shares]
    c_mounts = [_mount(ng, share_config) for ng in cluster.node_groups
                for share_config in cluster.shares or []]

    if not (ng_mounts or c_mounts):
        return

    ng_mounts_by_share_id = _group_mounts_by_share_id(ng_mounts)
    c_mounts_by_share_id = _group_mounts_by_share_id(c_mounts)

    all_share_ids = (set(ng_mounts_by_share_id.keys()) |
                     set(c_mounts_by_share_id.keys()))
    mounts_by_share_id = {
        share_id: c_mounts_by_share_id.get(share_id) or
        ng_mounts_by_share_id[share_id] for share_id in all_share_ids}

    all_mounts = itertools.chain(*mounts_by_share_id.values())
    mounts_by_ng_id = _group_mounts_by_ng_id(all_mounts)

    client = manila.client()
    handlers_by_share_id = {id: _ShareHandler.create_from_id(id, client)
                            for id in all_share_ids}

    for mounts in mounts_by_ng_id.values():
        node_group_shares = _NodeGroupShares(mounts[0].node_group)
        for mount in mounts:
            share_id = mount.share_config['id']
            node_group_shares.add_share(mount.share_config,
                                        handlers_by_share_id[share_id])
        node_group_shares.mount_shares_to_node_group()


def unmount_shares(cluster, unmount_share_list):
    """Unmounts all shares in unmount_share_list on the given cluster

    :param cluster: The cluster model.
    :param unmount_share_list: list of shares to unmount
    """
    client = manila.client()
    unmount_share_ids = (set(s['id'] for s in unmount_share_list))
    handlers_by_share_id = {id: _ShareHandler.create_from_id(id, client)
                            for id in unmount_share_ids}
    for share in unmount_share_list:
        for ng in cluster.node_groups:
            for instance in ng.instances:
                handlers_by_share_id[share['id']].unmount_from_instance(
                    instance.remote(), share)


_mount = collections.namedtuple('Mount', ['node_group', 'share_config'])


def _group_mounts(mounts, grouper):
    result = collections.defaultdict(list)
    for mount in mounts:
        result[grouper(mount)].append(mount)
    return result


def _group_mounts_by_share_id(mounts):
    return _group_mounts(mounts, lambda mount: mount.share_config['id'])


def _group_mounts_by_ng_id(mounts):
    return _group_mounts(mounts, lambda mount: mount.node_group['id'])


class _NodeGroupShares(object):
    """Organizes share mounting for a single node group."""

    _share = collections.namedtuple('Share', ['share_config', 'handler'])

    def __init__(self, node_group):
        self.node_group = node_group
        self.shares = []

    def add_share(self, share_config, handler):
        """Adds a share to mount; add all shares before mounting."""
        self.shares.append(self._share(share_config, handler))

    def mount_shares_to_node_group(self):
        """Mounts all configured shares to the node group."""
        for instance in self.node_group.instances:
            with context.set_current_instance_id(instance.instance_id):
                self._mount_shares_to_instance(instance)

    def _mount_shares_to_instance(self, instance):
        # Note: Additional iteration here is critical: based on
        # experimentation, failure to execute allow_access before spawning
        # the remote results in permission failure.
        for share in self.shares:
            share.handler.allow_access_to_instance(instance,
                                                   share.share_config)
        with instance.remote() as remote:
            share_types = set(type(share.handler) for share in self.shares)
            for share_type in share_types:
                share_type.setup_instance(remote)
            for share in self.shares:
                share.handler.mount_to_instance(remote, share.share_config)


@six.add_metaclass(abc.ABCMeta)
class _ShareHandler(object):
    """Handles mounting of a single share to any number of instances."""

    @classmethod
    def setup_instance(cls, remote):
        """Prepares an instance to mount this type of share."""
        pass

    @classmethod
    def create_from_id(cls, share_id, client):
        """Factory method for creation from a share_id of unknown type."""
        share = manila.get_share(client, share_id,
                                 raise_on_error=True)
        mounter_class = _share_types[share.share_proto]
        return mounter_class(share, client)

    def __init__(self, share, client):
        self.share = share
        self.client = client

    def allow_access_to_instance(self, instance, share_config):
        """Mounts a specific share to a specific instance."""
        access_level = self._get_access_level(share_config)
        accesses = list(filter(lambda x: (x.access_type == 'ip' and
                               x.access_to == instance.internal_ip),
                               self.share.access_list()))
        if accesses:
            access = accesses[0]
            if access.access_level not in ('ro', 'rw'):
                LOG.warning("Unknown permission level {access_level} on share "
                            "id {share_id} for ip {ip}. Leaving pre-existing "
                            "permissions.".format(
                                access_level=access.access_level,
                                share_id=self.share.id,
                                ip=instance.internal_ip))
            elif access.access_level == 'ro' and access_level == 'rw':
                self.share.deny(access.id)
                self.share.allow('ip', instance.internal_ip, access_level)
        else:
            self.share.allow('ip', instance.internal_ip, access_level)

    @abc.abstractmethod
    def mount_to_instance(self, remote, share_info):
        """Mounts the share to the instance as configured."""
        pass

    @abc.abstractmethod
    def unmount_from_instance(self, remote, share_info):
        """Unmounts the share from the instance."""
        pass

    def _get_access_level(self, share_config):
        return share_config.get('access_level', 'rw')

    def _default_mount(self):
        return '/mnt/{0}'.format(self.share.id)

    def _get_path(self, share_info):
        return share_info.get('path', self._default_mount())


class _NFSMounter(_ShareHandler):
    """Handles mounting of a single NFS share to any number of instances."""

    _DEBIAN_INSTALL = "dpkg -s nfs-common || apt-get -y install nfs-common"
    _REDHAT_INSTALL = "rpm -q nfs-utils || yum install -y nfs-utils"

    _NFS_CHECKS = {
        "centos": _REDHAT_INSTALL,
        "fedora": _REDHAT_INSTALL,
        "redhatenterpriseserver": _REDHAT_INSTALL,
        "redhat": _REDHAT_INSTALL,
        "ubuntu": _DEBIAN_INSTALL
    }

    _MKDIR_COMMAND = 'mkdir -p %s'
    _MOUNT_COMMAND = ("mount | grep '%(remote)s' | grep '%(local)s' | "
                      "grep nfs || mount -t nfs %(access_arg)s %(remote)s "
                      "%(local)s")
    _UNMOUNT_COMMAND = ("umount -f %s ")
    _RMDIR_COMMAND = 'rmdir %s'

    @classmethod
    def setup_instance(cls, remote):
        """Prepares an instance to mount this type of share."""
        distro = remote.get_os_distrib()
        if distro in cls._NFS_CHECKS:
            command = cls._NFS_CHECKS[distro]
            remote.execute_command(command, run_as_root=True)
        else:
            LOG.warning("Cannot verify installation of NFS mount tools for "
                        "unknown distro {distro}.".format(distro=distro))

    def mount_to_instance(self, remote, share_info):
        """Mounts the share to the instance as configured."""
        local_path = self._get_path(share_info)
        access_level = self._get_access_level(share_info)
        access_arg = '-w' if access_level == 'rw' else '-r'

        remote.execute_command(self._MKDIR_COMMAND % local_path,
                               run_as_root=True)
        mount_command = self._MOUNT_COMMAND % {
            "remote": self.share.export_location,
            "local": local_path,
            "access_arg": access_arg}
        remote.execute_command(mount_command, run_as_root=True)

    def unmount_from_instance(self, remote, share_info):
        """Unmounts the share from the instance."""
        local_path = self._get_path(share_info)

        unmount_command = self._UNMOUNT_COMMAND % local_path
        rmdir_command = self._RMDIR_COMMAND % local_path
        remote.execute_command(unmount_command, run_as_root=True)
        remote.execute_command(rmdir_command, run_as_root=True)


_share_types = {"NFS": _NFSMounter}
SUPPORTED_SHARE_TYPES = _share_types.keys()


def make_share_path(mount_point, path):
    return "{0}{1}".format(mount_point, path)


def default_mount(share_id):
    client = manila.client()
    return _ShareHandler.create_from_id(share_id, client)._default_mount()


def get_share_path(url, shares):
    # url example: 'manila://ManilaShare-uuid/path_to_file'
    url = six.moves.urllib.parse.urlparse(url)
    # using list() as a python2/3 workaround
    share_list = list(filter(lambda s: s['id'] == url.netloc, shares))
    if not share_list:
        # Share id is not in the share list, let the caller
        # determine a default path if possible
        path = None
    else:
        # We will always select the first one.  Let the
        # caller determine whether duplicates are okay
        mount_point = share_list[0].get('path', None)

        # Do this in two steps instead of passing the default
        # expression to get(), because it's a big side effect
        if mount_point is None:
            # The situation here is that the user specified a
            # share without a path, so the default mnt was used
            # during cluster provisioning.
            mount_point = default_mount(share_list[0]['id'])
        path = make_share_path(mount_point, url.path)
    return path



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils
import six
from six.moves.urllib import parse as urlparse

import sahara.exceptions as e
from sahara.i18n import _
from sahara.service.validations.edp import base as b
from sahara.utils import edp


DATA_TYPE_STRING = "string"
DATA_TYPE_NUMBER = "number"
DATA_TYPE_DATA_SOURCE = "data_source"

DATA_TYPES = [DATA_TYPE_STRING,
              DATA_TYPE_NUMBER,
              DATA_TYPE_DATA_SOURCE]
DEFAULT_DATA_TYPE = DATA_TYPE_STRING


INTERFACE_ARGUMENT_SCHEMA = {
    "type": ["array", "null"],
    "uniqueItems": True,
    "items": {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
                "minLength": 1
            },
            "description": {
                "type": ["string", "null"]
            },
            "mapping_type": {
                "type": "string",
                "enum": ["args", "configs", "params"]
            },
            "location": {
                "type": "string",
                "minLength": 1
            },
            "value_type": {
                "type": "string",
                "enum": DATA_TYPES,
                "default": "string"
            },
            "required": {
                "type": "boolean"
            },
            "default": {
                "type": ["string", "null"]
            }
        },
        "additionalProperties": False,
        "required": ["name", "mapping_type", "location", "required"]
    }
}


def _check_job_interface(data, interface):
    names = set(arg["name"] for arg in interface)
    if len(names) != len(interface):
        raise e.InvalidDataException(
            _("Name must be unique within the interface for any job."))

    mapping_types = set(arg["mapping_type"] for arg in interface)
    acceptable_types = edp.JOB_TYPES_ACCEPTABLE_CONFIGS[data["type"]]
    if any(m_type not in acceptable_types for m_type in mapping_types):
        args = {"mapping_types": str(list(acceptable_types)),
                "job_type": data["type"]}
        raise e.InvalidDataException(
            _("Only mapping types %(mapping_types)s are allowed for job type "
              "%(job_type)s.") % args)

    positional_args = [arg for arg in interface
                       if arg["mapping_type"] == "args"]
    if not all(six.text_type(arg["location"]).isnumeric()
               for arg in positional_args):
        raise e.InvalidDataException(
            _("Locations of positional arguments must be an unbroken integer "
              "sequence ascending from 0."))
    locations = set(int(arg["location"]) for arg in positional_args)
    if not all(i in locations for i in range(len(locations))):
        raise e.InvalidDataException(
            _("Locations of positional arguments must be an unbroken integer "
              "sequence ascending from 0."))

    not_required = (arg for arg in positional_args if not arg["required"])
    if not all(arg.get("default", None) for arg in not_required):
        raise e.InvalidDataException(
            _("Positional arguments must be given default values if they are "
              "not required."))

    mappings = ((arg["mapping_type"], arg["location"]) for arg in interface)
    if len(set(mappings)) != len(interface):
        raise e.InvalidDataException(
            _("The combination of mapping type and location must be unique "
              "within the interface for any job."))

    for arg in interface:
        if "value_type" not in arg:
            arg["value_type"] = DEFAULT_DATA_TYPE
        default = arg.get("default", None)
        if default is not None:
            _validate_value(arg["value_type"], default)


def check_job_interface(data, **kwargs):
    interface = data.get("interface", [])
    if interface:
        _check_job_interface(data, interface)


def _validate_data_source(value):
    if uuidutils.is_uuid_like(value):
        b.check_data_source_exists(value)
    else:
        if not urlparse.urlparse(value).scheme:
            raise e.InvalidDataException(
                _("Data source value '%s' is neither a valid data source ID "
                  "nor a valid URL.") % value)


def _validate_number(value):
    if not six.text_type(value).isnumeric():
        raise e.InvalidDataException(
            _("Value '%s' is not a valid number.") % value)


def _validate_string(value):
    if not isinstance(value, six.string_types):
        raise e.InvalidDataException(
            _("Value '%s' is not a valid string.") % value)


_value_type_validators = {
    DATA_TYPE_STRING: _validate_string,
    DATA_TYPE_NUMBER: _validate_number,
    DATA_TYPE_DATA_SOURCE: _validate_data_source
}


def _validate_value(type, value):
    _value_type_validators[type](value)


def check_execution_interface(data, job):
    job_int = {arg.name: arg for arg in job.interface}
    execution_int = data.get("interface", None)

    if not (job_int or execution_int):
        return
    if job_int and execution_int is None:
        raise e.InvalidDataException(
            _("An interface was specified with the template for this job. "
              "Please pass an interface map with this job (even if empty)."))

    execution_names = set(execution_int.keys())

    definition_names = set(job_int.keys())
    not_found_names = execution_names - definition_names
    if not_found_names:
        raise e.InvalidDataException(
            _("Argument names: %s were not found in the interface for this "
              "job.") % str(list(not_found_names)))

    required_names = {arg.name for arg in job.interface if arg.required}
    unset_names = required_names - execution_names
    if unset_names:
        raise e.InvalidDataException(_("Argument names: %s are required for "
                                       "this job.") % str(list(unset_names)))

    nonexistent = object()
    for name, value in six.iteritems(execution_int):
        arg = job_int[name]
        _validate_value(arg.value_type, value)
        if arg.mapping_type == "args":
            continue
        typed_configs = data.get("job_configs", {}).get(arg.mapping_type, {})
        config_value = typed_configs.get(arg.location, nonexistent)
        if config_value is not nonexistent and config_value != value:
            args = {"name": name,
                    "mapping_type": arg.mapping_type,
                    "location": arg.location}
            raise e.InvalidDataException(
                _("Argument '%(name)s' was passed both through the interface "
                  "and in location '%(mapping_type)s'.'%(location)s'. Please "
                  "pass this through either the interface or the "
                  "configuration maps, not both.") % args)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara import exceptions as ex
from sahara.i18n import _
from sahara.service.edp.utils import shares
from sahara.utils.openstack import manila


SHARE_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "id": {
                "type": "string",
                "format": "uuid"
            },
            "path": {
                "type": ["string", "null"]
            },
            "access_level": {
                "type": ["string", "null"],
                "enum": ["rw", "ro"],
                "default": "rw"
            }
        },
        "additionalProperties": False,
        "required": [
            "id"
        ]
    }
}


def check_shares(data):
    if not data:
        return

    paths = (share.get('path') for share in data)
    paths = [path for path in paths if path is not None]
    if len(paths) != len(set(paths)):
        raise ex.InvalidDataException(
            _('Multiple shares cannot be mounted to the same path.'))

    for path in paths:
        if not path.startswith('/') or '\x00' in path:
            raise ex.InvalidDataException(
                _('Paths must be absolute Linux paths starting with "/" '
                  'and may not contain nulls.'))

    client = manila.client()
    for share in data:
        manila_share = manila.get_share(client, share['id'])
        if not manila_share:
            raise ex.InvalidReferenceException(
                _("Requested share id %s does not exist.") % share['id'])

        share_type = manila_share.share_proto
        if share_type not in shares.SUPPORTED_SHARE_TYPES:
            raise ex.InvalidReferenceException(
                _("Requested share id %(id)s is of type %(type)s, which is "
                  "not supported by Sahara.")
                % {"id": share['id'], "type": share_type})



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
import datetime
import string

from oslo_config import cfg
from oslo_log import log as logging
import six

from sahara import conductor as c
from sahara import context
from sahara.i18n import _
from sahara.service import networks
from sahara.utils import cluster as cluster_utils
from sahara.utils import cluster_progress_ops as cpo
from sahara.utils import edp
from sahara.utils.openstack import base as b
from sahara.utils.openstack import images as sahara_images
from sahara.utils import poll_utils
from sahara.utils import remote

LOG = logging.getLogger(__name__)
conductor = c.API
CONF = cfg.CONF


@six.add_metaclass(abc.ABCMeta)
class Engine(object):

    @abc.abstractmethod
    def create_cluster(self, cluster):
        pass

    @abc.abstractmethod
    def scale_cluster(self, cluster, node_group_id_map):
        pass

    @abc.abstractmethod
    def shutdown_cluster(self, cluster, force):
        pass

    @abc.abstractmethod
    def rollback_cluster(self, cluster, reason):
        pass

    @abc.abstractmethod
    def get_type_and_version(self):
        """Returns engine type and version

         Result should be in the form 'type.major.minor'.
         """

    def get_node_group_image_username(self, node_group):
        image_id = node_group.get_image_id()
        return b.execute_with_retries(
            sahara_images.image_manager().get, image_id).username

    @poll_utils.poll_status('ips_assign_timeout', _("Assign IPs"), sleep=1)
    def _ips_assign(self, ips_assigned, cluster, instances):
        if not cluster_utils.check_cluster_exists(cluster):
            return True
        for instance in instances:
            if instance.id not in ips_assigned:
                with context.set_current_instance_id(instance.instance_id):
                    if networks.init_instances_ips(instance):
                        ips_assigned.add(instance.id)
                        cpo.add_successful_event(instance)
        return len(ips_assigned) == len(instances)

    def _await_networks(self, cluster, instances):
        if not instances:
            return

        cpo.add_provisioning_step(cluster.id, _("Assign IPs"), len(instances))

        ips_assigned = set()
        self._ips_assign(ips_assigned, cluster, instances)

        LOG.info("All instances have IPs assigned")

        cluster = conductor.cluster_get(context.ctx(), cluster)
        instances = cluster_utils.get_instances(cluster, ips_assigned)

        cpo.add_provisioning_step(
            cluster.id, _("Wait for instance accessibility"), len(instances))

        with context.ThreadGroup() as tg:
            for instance in instances:
                with context.set_current_instance_id(instance.instance_id):
                    tg.spawn("wait-for-ssh-%s" % instance.instance_name,
                             self._wait_until_accessible, instance)

        LOG.info("All instances are accessible")

    @poll_utils.poll_status(
        'wait_until_accessible', _("Wait for instance accessibility"),
        sleep=5)
    def _is_accessible(self, instance):
        if not cluster_utils.check_cluster_exists(instance.cluster):
            return True
        try:
            # check if ssh is accessible and cloud-init
            # script is finished generating authorized_keys
            exit_code, stdout = instance.remote().execute_command(
                "ls .ssh/authorized_keys", raise_when_error=False)

            if exit_code == 0:
                LOG.debug('Instance is accessible')
                return True
        except Exception as ex:
            ip_used = "internal_ip" if CONF.proxy_command and \
                CONF.proxy_command_use_internal_ip else "management_ip"
            LOG.debug("Can't login to node, IP: {ip}, reason {reason}"
                      .format(ip=getattr(instance, ip_used),
                              reason=ex))
            return False

        return False

    @cpo.event_wrapper(mark_successful_on_exit=True)
    def _wait_until_accessible(self, instance):
        self._is_accessible(instance)

    def _configure_instances(self, cluster):
        """Configure active instances.

        * generate /etc/hosts
        * change /etc/resolv.conf
        * setup passwordless login
        * etc.
        """
        cpo.add_provisioning_step(
            cluster.id, _("Configure instances"),
            cluster_utils.count_instances(cluster))

        with context.ThreadGroup() as tg:
            for node_group in cluster.node_groups:
                for instance in node_group.instances:
                    with context.set_current_instance_id(instance.instance_id):
                        tg.spawn("configure-instance-{}".format(
                            instance.instance_name),
                            self._configure_instance, instance, cluster
                        )

    @cpo.event_wrapper(mark_successful_on_exit=True)
    def _configure_instance(self, instance, cluster):
        self._configure_instance_etc_hosts(instance, cluster)
        if cluster.use_designate_feature():
            self._configure_instance_resolve_conf(instance)

    def _configure_instance_etc_hosts(self, instance, cluster):
        LOG.debug('Configuring "/etc/hosts" of instance.')
        hosts_file = cluster_utils.generate_etc_hosts(cluster)
        hostname = instance.fqdn()
        with instance.remote() as r:
            r.write_file_to('etc-hosts', hosts_file)
            r.write_file_to('etc-hostname', hostname)
            r.execute_command('sudo hostname %s' % hostname)
            r.execute_command('sudo cp etc-hosts /etc/hosts')
            r.execute_command('sudo cp etc-hostname /etc/hostname')
            r.execute_command('sudo rm etc-hosts  etc-hostname')
            r.execute_command('sudo usermod -s /bin/bash $USER')

    def _configure_instance_resolve_conf(self, instance):
        LOG.debug('Setting up those name servers from sahara.conf '
                  'which are lacked in the /etc/resolv.conf.')
        with instance.remote() as r:
            code, curr_resolv_conf = r.execute_command('cat /etc/resolv.conf')
            diff = cluster_utils.generate_resolv_conf_diff(curr_resolv_conf)
            if diff.strip():
                position = curr_resolv_conf.find('nameserver')
                if position == -1:
                    position = 0
                new_resolv_conf = "{}\n{}{}".format(
                    curr_resolv_conf[:position],
                    diff,
                    curr_resolv_conf[position:])
                r.write_file_to('resolv-conf', new_resolv_conf)
                r.execute_command('sudo mv resolv-conf /etc/resolv.conf')

    def _generate_user_data_script(self, node_group, instance_name):
        script = """#!/bin/bash
echo "${public_key}" >> ${user_home}/.ssh/authorized_keys\n
# ====== COMMENT OUT Defaults requiretty in /etc/sudoers ========
sed '/^Defaults    requiretty*/ s/^/#/' -i /etc/sudoers\n
"""

        script += remote.get_userdata_template()

        username = node_group.image_username

        if username == "root":
            user_home = "/root/"
        else:
            user_home = "/home/%s/" % username

        script_template = string.Template(script)

        return script_template.safe_substitute(
            public_key=node_group.cluster.management_public_key,
            user_home=user_home,
            instance_name=instance_name)

    # Deletion ops
    def _clean_job_executions(self, cluster):
        ctx = context.ctx()
        for je in conductor.job_execution_get_all(ctx, cluster_id=cluster.id):
            update = {"cluster_id": None}
            if not je.end_time:
                info = je.info.copy() if je.info else {}
                info['status'] = edp.JOB_STATUS_KILLED
                update.update({"info": info,
                               "end_time": datetime.datetime.now()})
            conductor.job_execution_update(ctx, je, update)

    def _remove_db_objects(self, cluster):
        ctx = context.ctx()
        cluster = conductor.cluster_get(ctx, cluster)
        instances = cluster_utils.get_instances(cluster)
        for inst in instances:
            conductor.instance_remove(ctx, inst)
